# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import os
import json
import tempfile
from copy import copy
from datetime import timedelta
from unittest import mock
from unittest.mock import MagicMock

from bson.objectid import ObjectId
from eve.utils import config, ParsedRequest
from eve.versioning import versioned_id_field

from apps.archive.archive import SOURCE as ARCHIVE
from apps.packages.package_service import PackageService
from apps.publish.content.common import BasePublishService
from apps.publish.content.publish import ArchivePublishService
from apps.publish.enqueue import enqueue_published, get_enqueue_service
from apps.publish.published_item import LAST_PUBLISHED_VERSION
from apps.prepopulate.app_populate import AppPopulateCommand
from superdesk import get_resource_service, get_backend
from superdesk.metadata.item import ITEM_STATE, CONTENT_STATE, ITEM_TYPE, CONTENT_TYPE
from superdesk.metadata.packages import RESIDREF
from superdesk.publish import init_app, publish_queue
from superdesk.publish.subscribers import SUBSCRIBER_TYPES
from superdesk.tests import TestCase
from superdesk.utc import utcnow
from apps.archive.common import ITEM_OPERATION
from celery.exceptions import SoftTimeLimitExceeded

ARCHIVE_PUBLISH = "archive_publish"
ARCHIVE_CORRECT = "archive_correct"
ARCHIVE_KILL = "archive_kill"

PUBLISH_QUEUE = "publish_queue"
PUBLISHED = "published"


@mock.patch("superdesk.publish.subscribers.SubscribersService.generate_sequence_number", lambda self, subscriber: 1)
class ArchivePublishTestCase(TestCase):
    def init_data(self):
        self.users = [{"_id": "1", "username": "admin"}]
        self.desks = [{"_id": ObjectId("123456789ABCDEF123456789"), "name": "desk1"}]
        self.products = [
            {"_id": "1", "name": "prod1", "geo_restrictions": "NSW", "email": "test@test.com"},
            {"_id": "2", "name": "prod2", "codes": "abc,def,"},
            {"_id": "3", "name": "prod3", "codes": "xyz"},
        ]
        self.subscribers = [
            {
                "_id": "1",
                "name": "sub1",
                "is_active": True,
                "subscriber_type": SUBSCRIBER_TYPES.WIRE,
                "media_type": "media",
                "sequence_num_settings": {"max": 10, "min": 1},
                "email": "test@test.com",
                "products": ["1"],
                "destinations": [
                    {
                        "name": "dest1",
                        "format": "nitf",
                        "delivery_type": "ftp",
                        "config": {"address": "127.0.0.1", "username": "test"},
                    }
                ],
            },
            {
                "_id": "2",
                "name": "sub2",
                "is_active": True,
                "subscriber_type": SUBSCRIBER_TYPES.WIRE,
                "media_type": "media",
                "sequence_num_settings": {"max": 10, "min": 1},
                "email": "test@test.com",
                "products": ["1"],
                "destinations": [
                    {
                        "name": "dest2",
                        "format": "nitf",
                        "delivery_type": "filecopy",
                        "config": {"address": "/share/copy"},
                    },
                    {
                        "name": "dest3",
                        "format": "nitf",
                        "delivery_type": "Email",
                        "config": {"recipients": "test@sourcefabric.org"},
                    },
                ],
            },
            {
                "_id": "3",
                "name": "sub3",
                "is_active": True,
                "subscriber_type": SUBSCRIBER_TYPES.DIGITAL,
                "media_type": "media",
                "sequence_num_settings": {"max": 10, "min": 1},
                "email": "test@test.com",
                "products": ["1"],
                "destinations": [
                    {
                        "name": "dest1",
                        "format": "nitf",
                        "delivery_type": "ftp",
                        "config": {"address": "127.0.0.1", "username": "test"},
                    }
                ],
            },
            {
                "_id": "4",
                "name": "sub4",
                "is_active": True,
                "subscriber_type": SUBSCRIBER_TYPES.WIRE,
                "media_type": "media",
                "sequence_num_settings": {"max": 10, "min": 1},
                "products": ["1"],
                "destinations": [
                    {
                        "name": "dest1",
                        "format": "nitf",
                        "delivery_type": "ftp",
                        "config": {"address": "127.0.0.1", "username": "test"},
                    }
                ],
            },
            {
                "_id": "5",
                "name": "sub5",
                "is_active": True,
                "subscriber_type": SUBSCRIBER_TYPES.ALL,
                "media_type": "media",
                "sequence_num_settings": {"max": 10, "min": 1},
                "email": "test@test.com",
                "codes": "xyz,  klm",
                "products": ["1", "2"],
                "destinations": [
                    {
                        "name": "dest1",
                        "format": "ninjs",
                        "delivery_type": "ftp",
                        "config": {"address": "127.0.0.1", "username": "test"},
                    }
                ],
            },
        ]

        self.articles = [
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a584-a7b402fed4f9",
                "_id": "1",
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                "last_version": 3,
                config.VERSION: 4,
                "body_html": "Test body",
                "anpa_category": [{"qcode": "A", "name": "Sport"}],
                "urgency": 4,
                "headline": "Two students missing",
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "ednote": "Andrew Marwood contributed to this article",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_STATE: CONTENT_STATE.PUBLISHED,
                "expiry": utcnow() + timedelta(minutes=20),
                "slugline": "story slugline",
                "unique_name": "#1",
                "operation": "publish",
            },
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a974-xy4532fe33f9",
                "_id": "2",
                "last_version": 3,
                config.VERSION: 4,
                "body_html": "Test body of the second article",
                "slugline": "story slugline",
                "urgency": 4,
                "anpa_category": [{"qcode": "A", "name": "Sport"}],
                "headline": "Another two students missing",
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "ednote": "Andrew Marwood contributed to this article",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                "expiry": utcnow() + timedelta(minutes=20),
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_STATE: CONTENT_STATE.PROGRESS,
                "publish_schedule": "2016-05-30T10:00:00+0000",
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                "unique_name": "#2",
                "operation": "publish",
            },
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a584-a7b402fed4fa",
                "_id": "3",
                "last_version": 3,
                config.VERSION: 4,
                "body_html": "Test body",
                "slugline": "story slugline",
                "urgency": 4,
                "anpa_category": [{"qcode": "A", "name": "Sport"}],
                "headline": "Two students missing killed",
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "ednote": "Andrew Marwood contributed to this article killed",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_STATE: CONTENT_STATE.KILLED,
                "expiry": utcnow() + timedelta(minutes=20),
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                "unique_name": "#3",
            },
            {
                "guid": "8",
                "_id": "8",
                "last_version": 3,
                config.VERSION: 4,
                "target_regions": [{"qcode": "NSW", "name": "New South Wales", "allow": True}],
                "body_html": "Take-1 body",
                "urgency": 4,
                "headline": "Take-1 headline",
                "abstract": "Abstract for take-1",
                "anpa_category": [{"qcode": "A", "name": "Sport"}],
                "pubstatus": "done",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "dateline": {"located": {"city": "Sydney"}},
                "slugline": "taking takes",
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_STATE: CONTENT_STATE.PROGRESS,
                "expiry": utcnow() + timedelta(minutes=20),
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                "unique_name": "#8",
            },
            {
                "_id": "9",
                "urgency": 3,
                "last_version": 3,
                config.VERSION: 4,
                "headline": "creator",
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_STATE: CONTENT_STATE.FETCHED,
            },
            {
                "guid": "tag:localhost:2015:69b961ab-a7b402fed4fb",
                "_id": "test_item_9",
                "last_version": 3,
                config.VERSION: 4,
                "body_html": "Student Crime. Police Missing.",
                "urgency": 4,
                "headline": "Police Missing",
                "abstract": "Police Missing",
                "anpa_category": [{"qcode": "A", "name": "Australian General News"}],
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "dateline": {"located": {"city": "Sydney"}},
                "slugline": "Police Missing",
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_STATE: CONTENT_STATE.PROGRESS,
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                "unique_name": "#9",
            },
            {
                "guid": "tag:localhost:10:10:10:2015:69b961ab-2816-4b8a-a584-a7b402fed4fc",
                "_id": "100",
                config.VERSION: 3,
                "task": {"user": "1", "desk": "123456789ABCDEF123456789"},
                ITEM_TYPE: CONTENT_TYPE.COMPOSITE,
                "groups": [
                    {"id": "root", "refs": [{"idRef": "main"}], "role": "grpRole:NEP"},
                    {
                        "id": "main",
                        "refs": [{"location": ARCHIVE, ITEM_TYPE: CONTENT_TYPE.COMPOSITE, RESIDREF: "6"}],
                        "role": "grpRole:main",
                    },
                ],
                "firstcreated": utcnow(),
                "expiry": utcnow() + timedelta(minutes=20),
                "unique_name": "#100",
                ITEM_STATE: CONTENT_STATE.PROGRESS,
            },
        ]

    def setUp(self):
        self.init_data()

        self.app.data.insert("users", self.users)
        self.app.data.insert("desks", self.desks)
        self.app.data.insert("products", self.products)
        self.app.data.insert("subscribers", self.subscribers)
        self.app.data.insert(ARCHIVE, self.articles)

        self.article_versions = self._init_article_versions()

        with tempfile.TemporaryDirectory() as tmp:
            json_data = [
                {"_id": "kill_text", "act": "kill", "type": "text", "schema": {"headline": {"type": "string"}}},
                {"_id": "publish_text", "act": "publish", "type": "text", "schema": {}},
                {"_id": "correct_text", "act": "correct", "type": "text", "schema": {}},
                {"_id": "publish_composite", "act": "publish", "type": "composite", "schema": {}},
            ]

            filename = os.path.join(tmp, "validators.json")
            with open(filename, "w") as file:
                json.dump(json_data, file)

            init_app(self.app)
            AppPopulateCommand().run(filename)

        self.app.media.url_for_media = MagicMock(return_value="url_for_media")
        self._put = self.app.media.put
        self.app.media.put = MagicMock(return_value="media_id")

    def tearDown(self):
        self.app.media.put = self._put

    def _init_article_versions(self):
        resource_def = self.app.config["DOMAIN"]["archive_versions"]
        version_id = versioned_id_field(resource_def)
        return [
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a584-a7b402fed4f9",
                version_id: "1",
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                config.VERSION: 1,
                "urgency": 4,
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                ITEM_STATE: CONTENT_STATE.DRAFT,
                "expiry": utcnow() + timedelta(minutes=20),
                "unique_name": "#8",
            },
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a584-a7b402fed4f9",
                version_id: "1",
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                config.VERSION: 2,
                "urgency": 4,
                "headline": "Two students missing",
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                ITEM_STATE: CONTENT_STATE.SUBMITTED,
                "expiry": utcnow() + timedelta(minutes=20),
                "unique_name": "#8",
            },
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a584-a7b402fed4f9",
                version_id: "1",
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                config.VERSION: 3,
                "urgency": 4,
                "headline": "Two students missing",
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "ednote": "Andrew Marwood contributed to this article",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                ITEM_STATE: CONTENT_STATE.PROGRESS,
                "expiry": utcnow() + timedelta(minutes=20),
                "unique_name": "#8",
            },
            {
                "guid": "tag:localhost:2015:69b961ab-2816-4b8a-a584-a7b402fed4f9",
                version_id: "1",
                ITEM_TYPE: CONTENT_TYPE.TEXT,
                config.VERSION: 4,
                "body_html": "Test body",
                "urgency": 4,
                "headline": "Two students missing",
                "pubstatus": "usable",
                "firstcreated": utcnow(),
                "byline": "By Alan Karben",
                "ednote": "Andrew Marwood contributed to this article",
                "dateline": {"located": {"city": "Sydney"}},
                "keywords": ["Student", "Crime", "Police", "Missing"],
                "subject": [{"qcode": "17004000", "name": "Statistics"}, {"qcode": "04001002", "name": "Weather"}],
                ITEM_STATE: CONTENT_STATE.PROGRESS,
                "expiry": utcnow() + timedelta(minutes=20),
                "unique_name": "#8",
            },
        ]

    def _is_publish_queue_empty(self):
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(0, queue_items.count())

    def _add_content_filters(self, product, is_global=False):
        product["content_filter"] = {"filter_id": 1, "filter_type": "blocking"}
        self.app.data.insert(
            "filter_conditions", [{"_id": 1, "field": "headline", "operator": "like", "value": "tor", "name": "test-1"}]
        )
        self.app.data.insert(
            "filter_conditions", [{"_id": 2, "field": "urgency", "operator": "in", "value": "2", "name": "test-2"}]
        )
        self.app.data.insert(
            "filter_conditions",
            [{"_id": 3, "field": "headline", "operator": "endswith", "value": "tor", "name": "test-3"}],
        )
        self.app.data.insert(
            "filter_conditions", [{"_id": 4, "field": "urgency", "operator": "in", "value": "2,3,4", "name": "test-4"}]
        )

        get_resource_service("content_filters").post(
            [
                {
                    "_id": 1,
                    "name": "pf-1",
                    "is_global": is_global,
                    "content_filter": [{"expression": {"fc": [4, 3]}}, {"expression": {"fc": [1, 2]}}],
                }
            ]
        )

    def test_publish(self):
        doc = self.articles[3].copy()
        get_resource_service(ARCHIVE_PUBLISH).patch(id=doc["_id"], updates={ITEM_STATE: CONTENT_STATE.PUBLISHED})
        published_doc = get_resource_service(ARCHIVE).find_one(req=None, _id=doc["_id"])
        self.assertIsNotNone(published_doc)
        self.assertEqual(published_doc[config.VERSION], doc[config.VERSION] + 1)
        self.assertEqual(published_doc[ITEM_STATE], ArchivePublishService().published_state)

    def test_versions_across_collections_after_publish(self):
        self.app.data.insert("archive_versions", self.article_versions)

        # Publishing an Article
        doc = self.articles[3]
        original = doc.copy()

        published_version_number = original[config.VERSION] + 1
        get_resource_service(ARCHIVE_PUBLISH).patch(
            id=doc[config.ID_FIELD],
            updates={ITEM_STATE: CONTENT_STATE.PUBLISHED, config.VERSION: published_version_number},
        )

        article_in_production = get_resource_service(ARCHIVE).find_one(req=None, _id=original[config.ID_FIELD])
        self.assertIsNotNone(article_in_production)
        self.assertEqual(article_in_production[ITEM_STATE], CONTENT_STATE.PUBLISHED)
        self.assertEqual(article_in_production[config.VERSION], published_version_number)

        enqueue_published()

        lookup = {"item_id": original[config.ID_FIELD], "item_version": published_version_number}
        queue_items = list(get_resource_service(PUBLISH_QUEUE).get(req=None, lookup=lookup))
        assert len(queue_items) > 0, "Transmission Details are empty for published item %s" % original[config.ID_FIELD]

        lookup = {"item_id": original[config.ID_FIELD], config.VERSION: published_version_number}
        request = ParsedRequest()
        request.args = {"aggregations": 0}
        items_in_published_collection = list(get_resource_service(PUBLISHED).get(req=request, lookup=lookup))
        assert len(items_in_published_collection) > 0, (
            "Item not found in published collection %s" % original[config.ID_FIELD]
        )

    def test_queue_transmission_for_item_scheduled_future(self):
        self._is_publish_queue_empty()

        doc = copy(self.articles[5])
        doc["item_id"] = doc["_id"]
        schedule_date = utcnow() + timedelta(hours=2)
        updates = {"publish_schedule": schedule_date, "schedule_settings": {"utc_publish_schedule": schedule_date}}
        get_resource_service(ARCHIVE).patch(id=doc["_id"], updates=updates)
        get_resource_service(ARCHIVE_PUBLISH).patch(id=doc["_id"], updates=updates)
        enqueue_published()
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(0, queue_items.count())

    def test_queue_transmission_for_item_scheduled_elapsed(self):
        self._is_publish_queue_empty()

        doc = copy(self.articles[5])
        doc["item_id"] = doc["_id"]
        schedule_date = utcnow() + timedelta(minutes=10)
        updates = {"publish_schedule": schedule_date, "schedule_settings": {"utc_publish_schedule": schedule_date}}
        get_resource_service(ARCHIVE).patch(id=doc["_id"], updates=updates)
        get_resource_service(ARCHIVE_PUBLISH).patch(id=doc["_id"], updates=updates)
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(0, queue_items.count())
        schedule_in_past = utcnow() + timedelta(minutes=-10)
        get_resource_service(PUBLISHED).update_published_items(
            doc["_id"], "schedule_settings", {"utc_publish_schedule": schedule_in_past}
        )
        get_resource_service(PUBLISHED).update_published_items(doc["_id"], "publish_schedule", schedule_in_past)

        enqueue_published()
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(1, queue_items.count())

    def test_queue_transmission_for_digital_channels(self):
        self._is_publish_queue_empty()

        doc = copy(self.articles[1])
        doc["item_id"] = doc["_id"]

        service = get_enqueue_service(doc[ITEM_OPERATION])
        subscribers, subscriber_codes, associations = service.get_subscribers(doc, SUBSCRIBER_TYPES.DIGITAL)
        service.queue_transmission(doc, subscribers, subscriber_codes)

        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(1, queue_items.count())
        expected_subscribers = ["5"]
        for item in queue_items:
            self.assertIn(item["subscriber_id"], expected_subscribers, "item {}".format(item))

    def test_queue_transmission_for_wire_channels_with_codes(self):
        self._is_publish_queue_empty()

        doc = copy(self.articles[1])
        doc["item_id"] = doc["_id"]

        service = get_enqueue_service(doc[ITEM_OPERATION])
        subscribers, subscriber_codes, associations = service.get_subscribers(doc, SUBSCRIBER_TYPES.WIRE)
        service.queue_transmission(doc, subscribers, subscriber_codes)
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]

        self.assertEqual(1, queue_items.count())
        expected_subscribers = ["5"]
        for item in queue_items:
            self.assertIn(item["subscriber_id"], expected_subscribers, "item {}".format(item))
            if item["subscriber_id"] == "5":
                self.assertEqual(4, len(item["codes"]))
                self.assertIn("def", item["codes"])
                self.assertIn("abc", item["codes"])
                self.assertIn("xyz", item["codes"])
                self.assertIn("klm", item["codes"])

    def test_get_subscribers_without_product(self):
        doc = copy(self.articles[1])
        doc["item_id"] = doc["_id"]

        subscriber_service = get_resource_service("subscribers")

        for sub in self.subscribers:
            sub.pop("products", None)
            subscriber_service.delete({"_id": sub["_id"]})

        subscriber_service.post(self.subscribers)

        service = get_enqueue_service(doc[ITEM_OPERATION])
        subscribers, subscriber_codes, associations = service.get_subscribers(doc, SUBSCRIBER_TYPES.WIRE)

        self.assertEqual(0, len(subscribers))
        self.assertDictEqual({}, subscriber_codes)

    def test_queue_transmission_wrong_article_type_fails(self):
        self._is_publish_queue_empty()

        doc = copy(self.articles[0])
        doc["item_id"] = doc["_id"]
        doc[ITEM_TYPE] = CONTENT_TYPE.PICTURE
        service = get_enqueue_service(doc[ITEM_OPERATION])

        subscribers, subscriber_codes, associations = service.get_subscribers(doc, SUBSCRIBER_TYPES.DIGITAL)
        no_formatters, queued = get_enqueue_service("publish").queue_transmission(doc, subscribers, subscriber_codes)
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(1, queue_items.count())
        self.assertEqual(0, len(no_formatters))
        self.assertTrue(queued)

        subscribers, subscriber_codes, associations = service.get_subscribers(doc, SUBSCRIBER_TYPES.WIRE)
        no_formatters, queued = get_enqueue_service("publish").queue_transmission(doc, subscribers)
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(2, queue_items.count())
        self.assertEqual(0, len(no_formatters))
        self.assertTrue(queued)

    def test_delete_from_queue_by_article_id(self):
        self._is_publish_queue_empty()

        doc = copy(self.articles[3])
        doc["item_id"] = doc["_id"]

        archive_publish = get_resource_service(ARCHIVE_PUBLISH)
        archive_publish.patch(id=doc["_id"], updates={ITEM_STATE: CONTENT_STATE.PUBLISHED})

        enqueue_published()
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(6, queue_items.count())

        # this will delete queue transmission for the wire article
        publish_queue.PublishQueueService(PUBLISH_QUEUE, get_backend()).delete_by_article_id(doc["_id"])
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(0, queue_items.count())

    def test_conform_target_regions(self):
        doc = {"headline": "test"}
        product = {"geo_restrictions": "QLD"}
        self.assertFalse(get_enqueue_service("publish").conforms_product_targets(product, doc))
        doc = {"headline": "test", "target_regions": []}
        self.assertFalse(get_enqueue_service("publish").conforms_product_targets(product, doc))
        doc = {"headline": "test", "target_regions": [{"qcode": "VIC", "name": "Victoria", "allow": True}]}
        self.assertFalse(get_enqueue_service("publish").conforms_product_targets(product, doc))
        doc = {"headline": "test", "target_regions": [{"qcode": "VIC", "name": "Victoria", "allow": False}]}
        self.assertTrue(get_enqueue_service("publish").conforms_product_targets(product, doc))
        doc = {"headline": "test", "target_regions": [{"qcode": "QLD", "name": "Queensland", "allow": True}]}
        self.assertTrue(get_enqueue_service("publish").conforms_product_targets(product, doc))
        doc = {"headline": "test", "target_regions": [{"qcode": "QLD", "name": "Queensland", "allow": False}]}
        self.assertFalse(get_enqueue_service("publish").conforms_product_targets(product, doc))

    def test_conform_target_subscribers(self):
        doc = {"headline": "test"}
        subscriber = {"_id": 1}
        self.assertTupleEqual(
            (True, False), get_enqueue_service("publish").conforms_subscriber_targets(subscriber, doc)
        )
        doc = {"headline": "test", "target_subscribers": []}
        self.assertTupleEqual(
            (True, False), get_enqueue_service("publish").conforms_subscriber_targets(subscriber, doc)
        )
        doc = {"headline": "test", "target_subscribers": [{"_id": 2}]}
        self.assertTupleEqual(
            (False, False), get_enqueue_service("publish").conforms_subscriber_targets(subscriber, doc)
        )
        doc = {"headline": "test", "target_subscribers": [{"_id": 1}]}
        self.assertTupleEqual((True, True), get_enqueue_service("publish").conforms_subscriber_targets(subscriber, doc))
        doc = {"headline": "test", "target_subscribers": [{"_id": 2}], "target_regions": [{"name": "Victoria"}]}
        self.assertTupleEqual(
            (True, False), get_enqueue_service("publish").conforms_subscriber_targets(subscriber, doc)
        )

    def test_can_publish_article(self):
        product = self.products[0]
        self._add_content_filters(product, is_global=False)

        service = get_enqueue_service("publish")
        can_it = service.conforms_content_filter(product, self.articles[4])
        self.assertFalse(can_it)
        product["content_filter"]["filter_type"] = "permitting"

        can_it = service.conforms_content_filter(product, self.articles[4])
        self.assertTrue(can_it)
        product.pop("content_filter")

    def test_can_publish_article_with_global_filters(self):
        subscriber = self.subscribers[0]
        product = self.products[0]
        self._add_content_filters(product, is_global=True)

        service = get_resource_service("content_filters")
        req = ParsedRequest()
        req.args = {"is_global": True}
        global_filters = list(service.get(req=req, lookup=None))
        enqueue_service = get_enqueue_service("publish")
        enqueue_service.conforms_global_filter(global_filters, self.articles[4])
        can_it = enqueue_service.conforms_subscriber_global_filter(subscriber, global_filters)
        self.assertFalse(can_it)

        subscriber["global_filters"] = {"1": False}
        can_it = enqueue_service.conforms_subscriber_global_filter(subscriber, global_filters)
        self.assertTrue(can_it)

        product.pop("content_filter")

    def test_is_targeted(self):
        doc = {"headline": "test"}
        self.assertFalse(BasePublishService().is_targeted(doc))
        doc = {"headline": "test", "target_regions": []}
        self.assertFalse(BasePublishService().is_targeted(doc))
        doc = {"headline": "test", "target_regions": [{"qcode": "NSW"}]}
        self.assertTrue(BasePublishService().is_targeted(doc))
        doc = {"headline": "test", "target_regions": [], "target_types": []}
        self.assertFalse(BasePublishService().is_targeted(doc))
        doc = {"headline": "test", "target_regions": [], "target_types": [{"qcode": "digital"}]}
        self.assertTrue(BasePublishService().is_targeted(doc))

    def test_targeted_for_includes_digital_subscribers(self):
        updates = {"target_regions": [{"qcode": "NSW", "name": "New South Wales", "allow": True}]}
        doc_id = self.articles[5][config.ID_FIELD]
        get_resource_service(ARCHIVE).patch(id=doc_id, updates=updates)

        get_resource_service(ARCHIVE_PUBLISH).patch(id=doc_id, updates={ITEM_STATE: CONTENT_STATE.PUBLISHED})
        enqueue_published()
        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(6, queue_items.count())
        expected_subscribers = ["1", "2", "3", "4", "5"]
        for item in queue_items:
            self.assertIn(item["subscriber_id"], expected_subscribers, "item {}".format(item))

    def test_maintain_latest_version_for_published(self):
        def get_publish_items(item_id, last_version):
            query = {
                "query": {
                    "filtered": {
                        "filter": {
                            "and": [{"term": {"item_id": item_id}}, {"term": {LAST_PUBLISHED_VERSION: last_version}}]
                        }
                    }
                }
            }
            request = ParsedRequest()
            request.args = {"source": json.dumps(query), "aggregations": 0}
            return self.app.data.find(PUBLISHED, req=request, lookup=None)[0]

        get_resource_service(ARCHIVE).patch(id=self.articles[1][config.ID_FIELD], updates={"publish_schedule": None})

        doc = get_resource_service(ARCHIVE).find_one(req=None, _id=self.articles[1][config.ID_FIELD])
        get_resource_service(ARCHIVE_PUBLISH).patch(
            id=doc[config.ID_FIELD], updates={ITEM_STATE: CONTENT_STATE.PUBLISHED}
        )

        enqueue_published()

        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(1, queue_items.count())
        request = ParsedRequest()
        request.args = {"aggregations": 0}
        published_items = self.app.data.find(PUBLISHED, request, None)[0]
        self.assertEqual(1, published_items.count())
        published_doc = next((item for item in published_items if item.get("item_id") == doc[config.ID_FIELD]), None)
        self.assertEqual(published_doc[LAST_PUBLISHED_VERSION], True)

        get_resource_service(ARCHIVE_CORRECT).patch(
            id=doc[config.ID_FIELD], updates={ITEM_STATE: CONTENT_STATE.CORRECTED}
        )

        enqueue_published()

        queue_items = self.app.data.find(PUBLISH_QUEUE, None, None)[0]
        self.assertEqual(2, queue_items.count())
        published_items = self.app.data.find(PUBLISHED, request, None)[0]
        self.assertEqual(2, published_items.count())
        last_published = get_publish_items(published_doc["item_id"], True)
        self.assertEqual(1, last_published.count())

    def test_added_removed_in_a_package(self):
        package = {
            "groups": [
                {"id": "root", "refs": [{"idRef": "main"}], "role": "grpRole:NEP"},
                {
                    "id": "main",
                    "refs": [
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "123",
                            "headline": "item-1 headline",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "123",
                        },
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "456",
                            "headline": "item-2 headline",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "456",
                        },
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "789",
                            "headline": "item-3 headline",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "789",
                        },
                    ],
                    "role": "grpRole:main",
                },
            ],
            "task": {
                "user": "#CONTEXT_USER_ID#",
                "status": "todo",
                "stage": "#desks.incoming_stage#",
                "desk": "#desks._id#",
            },
            "guid": "compositeitem",
            "headline": "test package",
            "state": "submitted",
            "type": "composite",
        }

        updates = {
            "groups": [
                {"id": "root", "refs": [{"idRef": "main"}], "role": "grpRole:NEP"},
                {
                    "id": "main",
                    "refs": [
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "123",
                            "headline": "item-1 headline",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "123",
                        },
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "555",
                            "headline": "item-2 headline",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "555",
                        },
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "456",
                            "headline": "item-2 headline",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "456",
                        },
                    ],
                    "role": "grpRole:main",
                },
            ],
            "task": {
                "user": "#CONTEXT_USER_ID#",
                "status": "todo",
                "stage": "#desks.incoming_stage#",
                "desk": "#desks._id#",
            },
            "guid": "compositeitem",
            "headline": "test package",
            "state": "submitted",
            "type": "composite",
        }

        items = PackageService().get_residrefs(package)
        removed_items, added_items = ArchivePublishService()._get_changed_items(items, updates)
        self.assertEqual(len(removed_items), 1)
        self.assertEqual(len(added_items), 1)

    def test_get_changed_items_no_item_found(self):
        # dummy publishing so that elastic mappings are created.
        doc = self.articles[3].copy()
        get_resource_service(ARCHIVE_PUBLISH).patch(id=doc["_id"], updates={ITEM_STATE: CONTENT_STATE.PUBLISHED})
        removed_items, added_items = get_enqueue_service("publish")._get_changed_items({}, {"item_id": "test"})
        self.assertEqual(len(removed_items), 0)
        self.assertEqual(len(added_items), 0)

    def test_reload_filters_if_updated(self):
        self.app.data.insert(
            "vocabularies",
            [
                {"_id": "categories", "items": []},
                {"_id": "urgency", "items": []},
                {"_id": "priority", "items": []},
                {"_id": "type", "items": []},
                {"_id": "genre", "items": []},
                {"_id": "place", "items": []},
            ],
        )
        product = self.products[0]
        product["content_filter"] = {"filter_id": 1, "filter_type": "blocking"}
        self.app.data.insert(
            "filter_conditions",
            [
                {
                    "_id": "1",
                    "field": "headline",
                    "operator": "like",
                    "value": "tor",
                    "name": "test-1",
                    "_updated": utcnow() - timedelta(days=10),
                }
            ],
        )
        get_enqueue_service("publish")
        get_resource_service("filter_conditions").patch("1", updates={"name": "test-1 updated"})
        service2 = get_enqueue_service("publish")
        self.assertGreater(service2.filters["latest_filter_conditions"], utcnow() - timedelta(seconds=10))


class TimeoutTest(TestCase):

    published_items = [
        {
            "_id": ObjectId("58006b8d1d41c88eace5179d"),
            "item_id": "1",
            "_created": utcnow(),
            "_updated": utcnow(),
            "queue_state": "pending",
            "state": "published",
            "operation": "publish",
        }
    ]

    def setUp(self):
        with self.app.app_context():
            init_app(self.app)

    @mock.patch("apps.publish.enqueue.get_enqueue_service", side_effect=SoftTimeLimitExceeded())
    def test_soft_timeout_gets_re_queued(self, mock):
        self.app.data.insert("published", self.published_items)
        enqueue_published()
        published = self.app.data.find(PUBLISHED, None, None)[0]
        self.assertTrue(published[0].get("queue_state"), "pending")



import superdesk
from flask import current_app as app
from superdesk.media.renditions import generate_renditions, get_renditions_spec
from apps.picture_crop import get_file


class PictureRenditionsService(superdesk.Service):
    """

    Create the renditions for the given `original` picture.

    """

    def create(self, docs, **kwargs):
        ids = []
        for doc in docs:
            item = doc.pop("item")
            orig = item["renditions"]["original"]
            orig_file = get_file(orig, item)
            no_custom_crops = doc.get("no_custom_crops", False)
            rendition_spec = get_renditions_spec(no_custom_crops=no_custom_crops)
            inserted = []
            mimetype = item.get("mimetype", orig.get("mimetype", "/"))
            media_type, content_type = mimetype.split("/")
            renditions = generate_renditions(
                orig_file, orig["media"], inserted, media_type, mimetype, rendition_spec, app.media.url_for_media
            )

            doc["renditions"] = renditions
            ids.append(item["_id"])

        return ids


class PictureRenditionsResource(superdesk.Resource):
    item_methods = []
    resource_methods = ["POST"]
    privileges = {"POST": "archive"}
    schema = {
        "item": {"type": "dict", "required": True},
        "no_custom_crops": {"type": "boolean"},
        "renditions": {
            "type": "dict",
            "schema": {},
            "allow_unknown": True,
        },
    }


def init_app(app):
    superdesk.register_resource("picture_renditions", PictureRenditionsResource, PictureRenditionsService, "archive")



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import superdesk
from superdesk.resource import Resource
from superdesk.services import BaseService
from apps.auth.errors import CredentialsAuthError
from superdesk import get_resource_service


logger = logging.getLogger(__name__)


class ChangePasswordResource(Resource):
    schema = {
        "username": {"type": "string", "required": True},
        "old_password": {"type": "string", "required": True},
        "new_password": {"type": "string", "required": True},
    }
    public_methods = ["POST"]
    resource_methods = ["POST"]
    item_methods = []


class ChangePasswordService(BaseService):
    def create(self, docs, **kwargs):
        for doc in docs:
            username = doc["username"]
            credentials = {"username": username, "password": doc["old_password"]}
            try:
                get_resource_service("auth_db").authenticate(credentials, True)
            except Exception as e:
                raise CredentialsAuthError(credentials=credentials, error=e)

            user = superdesk.get_resource_service("users").find_one(req=None, username=username)
            superdesk.get_resource_service("users").update_password(user["_id"], doc["new_password"])
            del doc["old_password"]
            del doc["new_password"]

            # return etag for further user updates
            user = superdesk.get_resource_service("users").find_one(req=None, _id=user["_id"])
            doc["_etag"] = user["_etag"]

            return [user["_id"]]



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2018 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging

from copy import deepcopy
from apps.tasks import send_to
from superdesk import register_resource, get_resource_service, privilege
from superdesk.services import Service
from superdesk.resource import Resource
from superdesk.errors import StopDuplication
from superdesk.signals import item_published, item_routed
from superdesk.metadata.item import PUBLISH_SCHEDULE, SCHEDULE_SETTINGS


NAME = "internal_destinations"

logger = logging.getLogger(__name__)


class InternalDestinationsResource(Resource):
    schema = {
        "name": {"type": "string", "required": True},
        "is_active": {"type": "boolean", "required": True},
        "filter": Resource.rel("content_filters", nullable=True),
        "desk": Resource.rel("desks", nullable=False, required=True),
        "stage": Resource.rel("stages", nullable=True),
        "macro": {"type": "string", "nullable": True},
        "send_after_schedule": {"type": "boolean"},
    }

    privileges = {"POST": "internal_destinations", "PATCH": "internal_destinations", "DELETE": "internal_destinations"}


class InternalDestinationsService(Service):
    pass


def handle_item_published(sender, item, **extra):
    macros_service = get_resource_service("macros")
    archive_service = get_resource_service("archive")
    filters_service = get_resource_service("content_filters")
    destinations_service = get_resource_service(NAME)

    for dest in destinations_service.get(req=None, lookup={"is_active": True}):
        if dest.get("desk") == item.get("task").get("desk"):
            # item desk and internal destination are same then don't execute
            continue

        if dest.get("filter"):
            content_filter = filters_service.find_one(req=None, _id=dest["filter"])
            if not content_filter:  # error state sort of, not sure what to do
                continue
            if not filters_service.does_match(content_filter, item):
                continue

        if dest.get("send_after_schedule", False) and item.get("state") != "published":
            # if send_after_schedule is set to True and item state is other than published
            # then don't execute
            continue
        elif item.get("state") == "published":
            item[PUBLISH_SCHEDULE] = None
            item[SCHEDULE_SETTINGS] = {}

        new_item = deepcopy(item)
        send_to(new_item, desk_id=dest["desk"], stage_id=dest.get("stage"))

        if dest.get("macro"):
            macro = macros_service.get_macro_by_name(dest["macro"])
            if not macro:
                logger.warning("macro %s not found for internal destination %s", dest["macro"], dest["name"])
            else:
                try:
                    macro["callback"](
                        new_item,
                        dest_desk_id=dest.get("desk"),
                        dest_stage_id=dest.get("stage"),
                    )
                except StopDuplication:
                    continue

        extra_fields = [PUBLISH_SCHEDULE, SCHEDULE_SETTINGS]
        next_id = archive_service.duplicate_item(new_item, state="routed", extra_fields=extra_fields)
        next_item = archive_service.find_one(req=None, _id=next_id)
        item_routed.send(sender, item=next_item)


def init_app(app):
    register_resource(NAME, InternalDestinationsResource, InternalDestinationsService, _app=app)

    privilege(name=NAME, label="Internal Destinations", description="User can manage internal destinations.")

    item_published.connect(handle_item_published)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from unittest import mock
from httmock import urlmatch, HTTMock
import os
from superdesk.tests import TestCase
from superdesk.io.feeding_services import APMediaFeedingService
from superdesk.io.feed_parsers import APMediaFeedParser

PREFIX = "test_superdesk_"
PROVIDER = {
    "_id": "test_provider",
    "config": {
        "api_url": "https://a.b.c/media/v/content/feed",
        "product_url": "https://api.ap.org/media/v/account/plans",
        "availableProducts": "12345",
        "apikey": "MYKEY",
        "productList": "",
    },
}


class APTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.setupMock(self)

    #        with self.app.app_context():
    #            vocab = [{}]
    #            self.app.data.insert('vocabularies', vocab)

    def setupMock(self, context):
        context.mock = HTTMock(
            *[self.feed_request],
            *[self.item_request],
            *[self.text_feed_request],
            *[self.text_item_request],
            *[self.nitf_item_request],
        )
        context.mock.__enter__()

    @urlmatch(scheme="https", netloc="a.b.c", path="/media/v/content/feed")
    def feed_request(self, url, request):
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", "ap_media_feed.json"))
        with open(fixture, "r") as f:
            feed_raw = f.read()
        return {"status_code": 200, "content": feed_raw}

    @urlmatch(scheme="https", netloc="a.b.c", path="/media/v/content/95e8bd71fed448cda1a0be35d8ddbd2f")
    def item_request(self, url, request):
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", "ap_media_item.json"))
        with open(fixture, "r") as f:
            feed_item = f.read()
        return {"status_code": 200, "content": feed_item}

    @mock.patch.object(APMediaFeedingService, "get_feed_parser")
    def test_feeding(self, get_feed_parser):
        get_feed_parser.return_value = APMediaFeedParser()
        provider = PROVIDER.copy()
        provider["config"]["api_url"] = "https://a.b.c/media/v/content/feed"
        service = APMediaFeedingService()
        service.provider = provider
        items = service._update(provider, {})[0]
        self.assertEqual(len(items), 1)
        self.assertEqual(items[0].get("headline"), "headline")

    @urlmatch(scheme="https", netloc="d.e.f", path="/media/v/content/feed")
    def text_feed_request(self, url, request):
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", "ap_media_text_feed.json"))
        with open(fixture, "r") as f:
            feed_raw = f.read()
        return {"status_code": 200, "content": feed_raw}

    @urlmatch(scheme="https", netloc="a.b.c", path="/media/v/content/5ee9b4b40ea5fe9c9ca60f890fb2e1d9")
    def text_item_request(self, url, request):
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", "ap_media_text_item.json"))
        with open(fixture, "r") as f:
            feed_item = f.read()
        return {"status_code": 200, "content": feed_item}

    @urlmatch(scheme="https", netloc="d.e.f", path="/media/v/content/5ee9b4b40ea5fe9c9ca60f890fb2e1d9.0/download")
    def nitf_item_request(self, url, request):
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", "ap_media_text_nitf.xml"))
        with open(fixture, "r") as f:
            feed_item = f.read()
        return {"status_code": 200, "content": feed_item}

    @mock.patch.object(APMediaFeedingService, "get_feed_parser")
    def test_text_feeding(self, get_feed_parser):
        get_feed_parser.return_value = APMediaFeedParser()
        provider = PROVIDER.copy()
        provider["config"]["api_url"] = "https://d.e.f/media/v/content/feed"
        provider["config"]["next_link"] = ""
        service = APMediaFeedingService()
        service.provider = provider
        items = service._update(provider, {})[0]
        self.assertEqual(len(items), 1)
        self.assertEqual(items[0].get("headline"), "BC-BBN--Top Ten")
        self.assertEqual(items[0].get("slugline"), "BBN--Top Ten")
        self.assertEqual(items[0].get("subject"), [{"name": "baseball", "qcode": "15007000"}])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from apps.auth import AuthResource
from .reset_password import ResetPasswordService, ResetPasswordResource, ActiveTokensResource
import superdesk
from .db import DbAuthService
from .commands import CreateUserCommand, HashUserPasswordsCommand  # noqa
from superdesk.services import BaseService
from apps.auth.db.change_password import ChangePasswordService, ChangePasswordResource


def init_app(app):
    endpoint_name = "auth_db"
    service = DbAuthService("auth", backend=superdesk.get_backend())
    AuthResource(endpoint_name, app=app, service=service)

    endpoint_name = "reset_user_password"
    service = ResetPasswordService(endpoint_name, backend=superdesk.get_backend())
    ResetPasswordResource(endpoint_name, app=app, service=service)

    endpoint_name = "change_user_password"
    service = ChangePasswordService(endpoint_name, backend=superdesk.get_backend())
    ChangePasswordResource(endpoint_name, app=app, service=service)

    endpoint_name = "active_tokens"
    service = BaseService(endpoint_name, backend=superdesk.get_backend())
    ActiveTokensResource(endpoint_name, app=app, service=service)


superdesk.intrinsic_privilege("auth_db", method=["DELETE"])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk
from superdesk.logging import logger
from superdesk.utc import utcnow
from superdesk.notification import push_notification
from apps.content import push_expired_notification
from superdesk.errors import ProviderError
from superdesk.lock import lock, unlock
from superdesk.io import get_feeding_service


class RemoveExpiredContent(superdesk.Command):
    """Remove stale data from ingest based on the provider settings.

    Example:
    ::

        $ python manage.py ingest:clean_expired
        $ python manage.py ingest:clean_expired --provider=aap

    """

    option_list = (superdesk.Option("--provider", "-p", dest="provider_name"),)

    def run(self, provider_name=None):
        providers = list(superdesk.get_resource_service("ingest_providers").get(req=None, lookup={}))
        self.remove_expired({"exclude": [str(p.get("_id")) for p in providers]})
        for provider in providers:
            if not provider_name or provider_name == provider.get("name"):
                self.remove_expired(provider)

    def remove_expired(self, provider):
        lock_name = "ingest:gc"

        if not lock(lock_name, expire=300):
            return

        try:

            remove_expired_data(provider)
            push_notification("ingest:cleaned")
        except Exception as err:
            logger.exception(err)
            raise ProviderError.expiredContentError(err, provider)
        finally:
            unlock(lock_name)


superdesk.command("ingest:clean_expired", RemoveExpiredContent())


def remove_expired_data(provider):
    """Remove expired data for provider"""
    logger.info("Removing expired content for provider: %s" % provider.get("_id", "Detached items"))

    try:
        feeding_service = get_feeding_service(provider["feeding_service"])
        ingest_collection = feeding_service.service if hasattr(feeding_service, "service") else "ingest"
    except KeyError:
        ingest_collection = "ingest"

    ingest_service = superdesk.get_resource_service(ingest_collection)

    items = get_expired_items(provider, ingest_collection)

    ids = [item["_id"] for item in items]
    items.rewind()
    file_ids = [
        rend.get("media")
        for item in items
        for rend in item.get("renditions", {}).values()
        if not item.get("archived") and rend.get("media")
    ]

    if ids:
        logger.info("Removing items %s" % ids)
        ingest_service.delete({"_id": {"$in": ids}})
        push_expired_notification(ids)

    for file_id in file_ids:
        logger.info("Deleting file: %s" % file_id)
        superdesk.app.media.delete(file_id)

    logger.info(
        "Removed expired content for provider: {0} count: {1}".format(provider.get("_id", "Detached items"), len(ids))
    )

    remove_expired_from_elastic(ingest_collection)


def remove_expired_from_elastic(ingest_collection):
    """Remove expired items from elastic which shouldn't be there anymore - expired before previous run."""
    ingest = superdesk.get_resource_service(ingest_collection)
    items = ingest.search({"filter": {"range": {"expiry": {"lt": "now-5m/m"}}}})
    if items.count():
        logger.warning("there are expired items in elastic (%d)" % (items.count(),))
        for item in items:
            logger.debug("doc only in elastic item=%s" % (item,))
            ingest.remove_from_search(item)


def get_expired_items(provider_id, ingest_collection):
    query_filter = get_query_for_expired_items(provider_id)
    return superdesk.get_resource_service(ingest_collection).get_from_mongo(lookup=query_filter, req=None)


def get_query_for_expired_items(provider):
    """Find all ingest items with given provider id and expiry is past

    :param dict provider: ingest provider
    :return str: mongo query
    """
    query = {"expiry": {"$lte": utcnow()}}

    if provider.get("_id"):
        query["ingest_provider"] = str(provider.get("_id"))

    if provider.get("exclude"):
        excluded = provider.get("exclude")
        query["ingest_provider"] = {"$nin": excluded}

    return query



import superdesk


def norvig_suggest(word, model):
    """Norvig's simple spell check.

    Modified not to return only best correction, but all of them sorted.
    """
    NWORDS = model
    alphabet = "abcdefghijklmnopqrstuvwxyz"
    suggestions = []

    def edits1(word):
        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes = [a + b[1:] for a, b in splits if b]
        transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b) > 1]
        replaces = [a + c + b[1:] for a, b in splits for c in alphabet if b]
        inserts = [a + c + b for a, b in splits for c in alphabet]

        return set(deletes + transposes + replaces + inserts)

    def known_edits2(word):
        return set(e2 for e1 in edits1(word) for e2 in edits1(e1) if e2 in NWORDS)

    def known(words):
        return set(w for w in words if w in NWORDS)

    def suggest(word):
        candidates = known([word]) or known(edits1(word))  # or known_edits2(word)
        return sorted(candidates, key=lambda item: NWORDS.get(item, 1), reverse=True)

    name_suggestions = suggest(word.capitalize())
    # check if word suggestion is a name as we store name as capitalized case in dictionary
    if len(name_suggestions) == 1:
        return name_suggestions

    # if not name suggestion find all the suggestions by converting the word to lowercase
    # and make sure to send lowercase as well as capitalized case suggestions for normal words
    for suggestion in suggest(word.lower()):
        suggestions.extend([suggestion, suggestion.capitalize()])

    return suggestions


class SpellcheckResource(superdesk.Resource):

    resource_methods = ["POST"]
    item_methods = []

    schema = {
        "word": {"type": "string", "required": True},
        "language_id": {"type": "string", "required": True},
        "corrections": {"type": "list"},
    }

    # you should be able to make edits
    privileges = {"POST": "archive"}


class SpellcheckService(superdesk.Service):
    def suggest(self, word, lang):
        """Suggest corrections for given word and language.

        :param word: word that is probably wrong
        :param lang: language code
        """
        model = superdesk.get_resource_service("dictionaries").get_model_for_lang(lang)
        return norvig_suggest(word, model)

    def create(self, docs, **kwargs):
        for doc in docs:
            doc["corrections"] = self.suggest(doc["word"], doc["language_id"])
        return [doc["word"] for doc in docs]



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license
from eve.utils import ParsedRequest

from apps.search import init_app
from superdesk.metadata.item import ITEM_STATE, CONTENT_STATE
from superdesk.tests import TestCase


def resource_listener(resource, docs):
    for doc in docs["_items"]:
        doc["_resource_listener"] = resource


def ingest_listener(docs):
    for doc in docs["_items"]:
        doc["_ingest_listener"] = 1


class SearchServiceTestCase(TestCase):
    def setUp(self):
        with self.app.app_context():
            self.app.data.insert("ingest", [{}])
            self.app.data.insert(
                "archive",
                [
                    {"_id": "456", "task": {"desk": 1}, ITEM_STATE: CONTENT_STATE.PROGRESS},
                    {"_id": "911", "task": {"desk": None, "user": "test"}, ITEM_STATE: CONTENT_STATE.PROGRESS},
                    {"_id": "123", "task": {"desk": 1}, ITEM_STATE: CONTENT_STATE.PUBLISHED},
                    {"_id": "786", "task": {"desk": 1}, ITEM_STATE: CONTENT_STATE.PUBLISHED},
                ],
            )
            self.app.data.insert(
                "published", [{"item_id": "786", "task": {"desk": 1}, ITEM_STATE: CONTENT_STATE.PUBLISHED}]
            )
            self.app.data.insert(
                "archived", [{"item_id": "123", "task": {"desk": 1}, ITEM_STATE: CONTENT_STATE.PUBLISHED}]
            )
            init_app(self.app)
            self.app.on_fetched_resource += resource_listener
            self.app.on_fetched_resource_ingest += ingest_listener

    def test_query_post_processing(self):
        with self.app.app_context():
            docs = self.app.data.find("search", None, None)[0]
            self.assertEquals(4, docs.count())

            ingest_docs = [doc for doc in docs if doc["_type"] == "ingest"]
            self.assertEquals("ingest", ingest_docs[0]["_resource_listener"])
            self.assertEquals(1, ingest_docs[0]["_ingest_listener"])

            archive_docs = [doc for doc in docs if doc["_type"] == "archive"]
            self.assertNotIn("_ingest_listener", archive_docs[0])

    def test_using_repo_request_attribute(self):
        with self.app.app_context():
            req = ParsedRequest()
            req.args = {"repo": "ingest"}
            docs = self.app.data.find("search", req, None)[0]
            self.assertEquals(1, docs.count())
            self.assertEquals("ingest", docs[0]["_type"])

    def test_it_filters_out_private_content(self):
        with self.app.app_context():
            self.app.data.insert("archive", [{"task": {"desk": None}}, {"task": {}}])
            cursor = self.app.data.find("search", None, None)[0]
            self.assertEquals(4, cursor.count())

    def test_it_includes_published_content(self):
        with self.app.app_context():
            cursor = self.app.data.find("search", None, None)[0]
            self.assertEquals(4, cursor.count())

    def test_it_excludes_published_content(self):
        with self.app.app_context():
            req = ParsedRequest()
            req.args = {"repo": "archive"}
            docs = self.app.data.find("search", req, None)[0]
            self.assertEquals(1, docs.count())



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


def update_fields(item, **kwargs):
    """Updates the abstract field"""
    item["abstract"] = "Abstract has been updated"
    return item


name = "update_fields"
label = "Update Fields"
shortcut = "w"
callback = update_fields
desks = ["POLITICS"]
access_type = "frontend"
action_type = "direct"



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

"""Superdesk Users"""

from superdesk.metadata.item import BYLINE, SIGN_OFF
from superdesk.resource import Resource


class UsersResource(Resource):
    def __init__(self, endpoint_name, app, service, endpoint_schema=None):
        self.readonly = True if app.config.get("LDAP_SERVER", None) else False

        self.additional_lookup = {"url": r'regex("[\w]+")', "field": "username"}

        self.schema = {
            "username": {
                "type": "string",
                "unique": True,
                "required": True,
                "minlength": 1,
                "username_pattern": True,
            },
            "password": {"type": "string", "minlength": 5},
            "password_changed_on": {"type": "datetime", "nullable": True},
            "first_name": {"type": "string", "readonly": self.readonly},
            "last_name": {"type": "string", "readonly": self.readonly},
            "display_name": {"type": "string"},
            "email": {"unique": True, "type": "email", "required": True, "coerce": lambda s: s.lower()},
            "phone": {"type": "string", "nullable": True},
            "job_title": {
                "type": "string",
                "required": False,
            },
            "biography": {
                "type": "string",
                "required": False,
                "nullable": True,
            },
            "facebook": {
                "type": "string",
                "required": False,
                "nullable": True,
            },
            "instagram": {
                "type": "string",
                "required": False,
                "nullable": True,
            },
            "twitter": {
                "type": "string",
                "required": False,
                "nullable": True,
                "twitter": True,
            },
            "jid": {
                "unique": True,
                "type": "string",
                "required": False,
            },
            "language": {"type": "string", "nullable": True},
            "user_info": {
                "type": "dict",
                "schema": {},
                "allow_unknown": True,
            },
            "picture_url": {"type": "string", "nullable": True},
            "avatar": Resource.rel("upload", embeddable=True, nullable=True),
            "avatar_renditions": {"type": "dict", "schema": {}},
            "role": Resource.rel("roles", True),
            "privileges": {
                "type": "dict",
                "schema": {},
                "allow_unknown": True,
            },
            "workspace": {
                "type": "dict",
                "schema": {},
                "allow_unknown": True,
            },
            "user_type": {"type": "string", "allowed": ["user", "administrator"], "default": "user"},
            "is_support": {"type": "boolean", "default": False},
            "is_author": {"type": "boolean", "default": True},
            "is_active": {"type": "boolean", "default": True},
            "is_enabled": {"type": "boolean", "default": True},
            "needs_activation": {"type": "boolean", "default": True},
            # Default desk of the user, which would be selected when logged-in.
            "desk": Resource.rel("desks", nullable=True),
            SIGN_OFF: {  # Used for putting a sign-off on the content when it's created/updated except kill
                "type": "string",
                "required": False,
                "nullable": True,
                "regex": "^[a-zA-Z0-9]+$",
            },
            BYLINE: {"type": "string", "required": False, "nullable": True},
            # list to hold invisible stages.
            # This field is updated under following scenario:
            # 1. stage visible flag is updated
            # 2. desk membership is modified
            # 3. new user is created
            "invisible_stages": {"type": "list", "required": False, "nullable": True},
            # If Slack notifications are configured and enabled for the user
            # the Slack username is stored here.
            "slack_username": {"type": "string", "required": False, "nullable": True},
            # The Slack user id is stored here, to avoid repeatedly having to look it up
            "slack_user_id": {"type": "string", "required": False, "nullable": True},
            "session_preferences": {
                "type": "dict",
                "schema": {},
                "allow_unknown": True,
            },
            "user_preferences": {
                "type": "dict",
                "schema": {},
                "allow_unknown": True,
            },
            "last_activity_at": {"type": "datetime", "nullable": True, "readonly": True},
        }

        self.extra_response_fields = [
            "display_name",
            "username",
            "email",
            "user_info",
            "picture_url",
            "avatar",
            "is_active",
            "is_enabled",
            "needs_activation",
            "desk",
            "last_activity_at",
        ]

        self.etag_ignore_fields = ["session_preferences", "_etag", "invisible_stages", "last_activity_at"]

        self.datasource = {
            "projection": {"password": 0, "session_preferences": 0},
            "default_sort": [("username", 1)],
        }

        self.mongo_indexes = {
            "username_1": ([("username", 1)], {"unique": True}),
            "first_name_1_last_name_-1": [("first_name", 1), ("last_name", -1)],
        }

        self.privileges = {"POST": "users", "DELETE": "users", "PATCH": "users"}
        super().__init__(endpoint_name, app=app, service=service, endpoint_schema=endpoint_schema)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license
import json

from eve.utils import ParsedRequest
from flask_babel import _

from superdesk import get_resource_service, app
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.errors import SuperdeskApiError


class ContentFilterTestResource(Resource):
    endpoint_name = "content_filter_tests"
    schema = {
        "filter_id": {"type": "string"},
        "article_id": {"type": "string"},
        "return_matching": {"type": "boolean"},
        "filter": {"type": "dict"},
        "match_results": {"type": "list"},
    }
    url = "content_filters/test"
    resource_methods = ["POST"]
    item_methods = []
    resource_title = endpoint_name
    privileges = {"POST": "content_filters"}


class ContentFilterTestService(BaseService):
    def create(self, docs, **kwargs):
        service = get_resource_service("content_filters")
        for doc in docs:
            filter_id = doc.get("filter_id")
            if filter_id:
                content_filter = service.find_one(req=None, _id=filter_id)
            else:
                content_filter = doc.get("filter")

            if not content_filter:
                raise SuperdeskApiError.badRequestError(_("Content filter not found"))

            if "article_id" in doc:
                article_id = doc.get("article_id")
                article = get_resource_service("archive").find_one(req=None, _id=article_id)
                if not article and "planning" in app.config.get("INSTALLED_APPS", []):
                    article = get_resource_service("planning").find_one(None, _id=article_id)
                if not article:
                    article = get_resource_service("ingest").find_one(req=None, _id=article_id)
                    if not article:
                        raise SuperdeskApiError.badRequestError(_("Article not found!"))
                try:
                    doc["match_results"] = service.does_match(content_filter, article)
                except Exception as ex:
                    raise SuperdeskApiError.badRequestError(
                        _("Error in testing article: {error}").format(error=str(ex))
                    )
            else:
                try:
                    if doc.get("return_matching", True):
                        query = service.build_elastic_query(content_filter)
                    else:
                        query = service.build_elastic_not_filter(content_filter)
                    query["sort"] = [{"versioncreated": "desc"}]
                    query["size"] = 200
                    req = ParsedRequest()
                    req.args = {"source": json.dumps(query)}
                    doc["match_results"] = list(get_resource_service("archive").get(req=req, lookup=None))
                except Exception as ex:
                    raise SuperdeskApiError.badRequestError(
                        _("Error in testing archive: {error}").format(error=str(ex))
                    )

        return [len(docs)]



import locale
import warnings
import superdesk

from flask import current_app as app, request
from flask_babel import _
from datetime import timedelta
from superdesk.utils import ListCursor
from superdesk.utc import utcnow
from superdesk.errors import SuperdeskApiError


SETTING_ENABLED = "ARCHIVE_AUTOCOMPLETE"
SETTING_DAYS = "ARCHIVE_AUTOCOMPLETE_DAYS"
SETTING_HOURS = "ARCHIVE_AUTOCOMPLETE_HOURS"


class AutocompleteResource(superdesk.Resource):
    item_methods = []
    resource_methods = ["GET"]
    schema = {
        "value": {"type": "string"},
    }


class AutocompleteService(superdesk.Service):

    allowed_fields = {
        "slugline",
    }

    def get(self, req, lookup):
        field = request.args.get("field", "slugline")
        language = request.args.get("language", app.config.get("DEFAULT_LANGUAGE", "en"))

        if not app.config.get(SETTING_ENABLED):
            raise SuperdeskApiError(_("Archive autocomplete is not enabled"), 404)

        if field not in self.allowed_fields:
            raise SuperdeskApiError(_("Field %(field)s is not allowed", field=field), 400)

        _filter = {
            "state": "published",
            "language": language,
            "versioncreated": {
                "$gte": utcnow()
                - timedelta(
                    days=app.config[SETTING_DAYS],
                    hours=app.config[SETTING_HOURS],
                )
            },
        }
        values = self.backend._backend("archive").driver.db["archive"].distinct(field, _filter)
        sorted_values = sorted(values, key=locale.strxfrm)
        docs = [{"value": value} for value in sorted_values]
        return ListCursor(docs)


def init_app(_app):
    _app.client_config.update({"archive_autocomplete": _app.config.get(SETTING_ENABLED, False)})

    if _app.config.get(SETTING_ENABLED) and not _app.config.get(SETTING_HOURS) and not _app.config.get(SETTING_DAYS):
        warnings.warn(
            "{} is enabled but both {} and {} are not set.".format(
                SETTING_ENABLED,
                SETTING_HOURS,
                SETTING_DAYS,
            )
        )

    superdesk.register_resource("archive_autocomplete", AutocompleteResource, AutocompleteService, _app=_app)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import logging

import superdesk

from .archive_copy import CopyService, CopyResource, init_app as archive_copy_init_app
from .archive_duplication import DuplicateService, DuplicateResource
from .archive_fetch import FetchResource, FetchService
from .archive_move import MoveResource, MoveService
from .archive_translate import TranslateService, TranslateResource
from flask_babel import lazy_gettext


logger = logging.getLogger(__name__)


def init_app(app):
    endpoint_name = "fetch"
    service = FetchService(endpoint_name, backend=superdesk.get_backend())
    FetchResource(endpoint_name, app=app, service=service)

    endpoint_name = "duplicate"
    service = DuplicateService(endpoint_name, backend=superdesk.get_backend())
    DuplicateResource(endpoint_name, app=app, service=service)

    endpoint_name = "translate"
    service = TranslateService(endpoint_name, backend=superdesk.get_backend())
    TranslateResource(endpoint_name, app=app, service=service)

    endpoint_name = "copy"
    service = CopyService(endpoint_name, backend=superdesk.get_backend())
    CopyResource(endpoint_name, app=app, service=service)

    endpoint_name = "move"
    service = MoveService(endpoint_name, backend=superdesk.get_backend())
    MoveResource(endpoint_name, app=app, service=service)

    superdesk.privilege(
        name="fetch", label=lazy_gettext("Fetch Content To a Desk"), description=lazy_gettext("Fetch Content to a Desk")
    )
    superdesk.privilege(
        name="move",
        label=lazy_gettext("Move Content to another desk"),
        description=lazy_gettext("Move Content to another desk"),
    )
    superdesk.privilege(
        name="duplicate",
        label=lazy_gettext("Duplicate Content within a Desk"),
        description=lazy_gettext("Duplicate Content within a Desk"),
    )
    superdesk.privilege(
        name="translate",
        label=lazy_gettext("Translate Content within a Desk"),
        description=lazy_gettext("Translate Content within a Desk"),
    )
    superdesk.privilege(
        name="send_to_personal",
        label=lazy_gettext("Send Content to Personal desk"),
        description=lazy_gettext("Send Content to Personal desk"),
    )

    superdesk.intrinsic_privilege("copy", method=["POST"])

    archive_copy_init_app(app)



import unittest

from superdesk.schema import Schema, StringField, IntegerField, ListField, DictField, SchemaField, NoneField


class TestSchema(Schema):

    string_field = StringField()

    integer_field = IntegerField()

    list_field = ListField()

    dict_field = DictField()

    empty_field = SchemaField()

    none_field = NoneField()


class SchemaTest(unittest.TestCase):
    def test_schema_iterable(self):
        schema = dict(TestSchema)
        self.assertIn("string_field", schema)
        self.assertEqual(
            schema["string_field"], {"type": "string", "required": False, "minlength": None, "maxlength": None}
        )
        self.assertEqual(schema["integer_field"], {"type": "integer", "required": False})
        self.assertEqual(
            schema["list_field"], {"type": "list", "required": False, "mandatory_in_list": None, "schema": None}
        )
        self.assertEqual(schema["dict_field"], {"type": "dict", "required": False})
        self.assertEqual(schema["empty_field"], {})
        self.assertEqual(None, schema["none_field"])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk

from superdesk.signals import item_published
from .vocabularies import VocabulariesResource, VocabulariesService, is_related_content
from .commands import UpdateVocabulariesInItemsCommand  # noqa
from .keywords import add_missing_keywords
from flask_babel import _


def init_app(app):
    endpoint_name = "vocabularies"
    service = VocabulariesService(endpoint_name, backend=superdesk.get_backend())
    VocabulariesResource(endpoint_name, app=app, service=service)

    superdesk.register_default_user_preference(
        "cvs:preferred_items",
        {
            "value": {},
            "category": _("cvs"),
            "label": _("Prefered CV items"),
        },
    )

    item_published.connect(add_missing_keywords)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import unittest
from .area_acre_to_metric import acre_to_metric


class TemperatureTestCase(unittest.TestCase):
    def test_f_to_c(self):
        text = (
            "Total area is 100.50 ac. for this land. "
            "Total area is 15.7 Acres for this land. "
            "Total area is 1 acre for this land. "
            "Total area is 1-16 acres for this land. "
            "Total area is 16.7-acres for this land. "
            "Total area is 16,500-acre for this land. "
        )
        item = {"body_html": text}
        res, diff = acre_to_metric(item)
        self.assertEqual(diff["100.50 ac"], "100.50 ac (40.7 ha)")
        self.assertEqual(diff["15.7 Acres"], "15.7 Acres (6.4 ha)")
        self.assertEqual(diff["1 acre"], "1 acre (4,047 square meter)")
        self.assertEqual(diff["1-16 acres"], "1-16 acres (0.4-6.5 ha)")
        self.assertEqual(diff["16.7-acres"], "16.7-acres (6.8 ha)")
        self.assertEqual(diff["16,500-acre"], "16,500-acre (6,677 ha)")
        self.assertEqual(res["body_html"], item["body_html"])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import unittest

from superdesk.macros.abstract_populator import populate


class AbstractPopulatorTestCase(unittest.TestCase):
    def setUp(self):
        self.body = (
            "Fountain of Time, or simply Time, is a sculpture by Lorado Taft, "
            "measuring 126 feet 10 inches (38.66 m) in length, situated at the "
            "western edge of the Midway Plaisance within Washington Park in "
            "Chicago, Illinois, in the United States. This location is in the "
            "Washington Park community area on Chicago's South Side. Inspired "
            'by Henry Austin Dobson\'s poem, "Paradox of Time", and with its '
            "100 figures passing before Father Time, the work was created as a "
            "monument to the first 100 years of peace between the United States "
            "and Great Britain, resulting from the Treaty of Ghent in 1814. "
            "Although the fountain's water began running in 1920, the sculpture "
            "was not dedicated to the city until 1922. The sculpture is a "
            "contributing structure to the Washington Park United States "
            "Registered Historic District, which is a National Register of "
            "Historic Places listing."
        )

    def test_populate(self):
        item = {"body_html": self.body}
        item = populate(item)
        self.assertEqual(item["abstract"], "Fountain of Time, or simply Time, is a sculpture by Lorado Taft,")

    def test_populate_without_body(self):
        item = {}
        item = populate(item)
        self.assertEqual(item["abstract"], "No body found to use for abstract...")



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from datetime import datetime, timedelta

from superdesk.tests import TestCase
from superdesk.io.commands.update_ingest import is_not_expired


class ItemExpiryTestCase(TestCase):
    def test_expiry_no_dateinfo(self):
        self.assertTrue(is_not_expired({}, None))

    def test_expiry_overflow(self):
        item = {"versioncreated": datetime.now()}
        delta = timedelta(minutes=999999999999)
        self.assertTrue(is_not_expired(item, delta))



# -*- coding: utf-8; -*-
# This file is part of Superdesk.
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license
#
# Author  : edouard
# Creation: 2016-06-10 11:10

from superdesk.commands.data_updates import BaseDataUpdate


class DataUpdate(BaseDataUpdate):
    """

    Data update for Pull Request #418

    SD-3487 As a user I want to be able to use a template in several desks
    https://github.com/superdesk/superdesk-core/pull/418


    Changes:

        1. `template_desks` added and is a list of desks, initialized with desk `template_desk`
        2. `template_stage` renamed by `schedule_stage`
        3. `template_desk` renamed by `schedule_desk`

    """

    resource = "content_templates"

    def forwards(self, mongodb_collection, mongodb_database):
        # new `template_desks` field must contain a list of desk id
        for template in mongodb_collection.find({}):
            if template.get("template_desk"):
                print(
                    mongodb_collection.update(
                        {"_id": template["_id"]}, {"$set": {"template_desks": [template.get("template_desk")]}}
                    )
                )
        # renames fields:
        #   - template_desk -> schedule_desk
        #   - template_stage -> schedule_stage
        print(
            mongodb_collection.update(
                {},
                {
                    "$rename": {"template_desk": "schedule_desk", "template_stage": "schedule_stage"},
                },
                upsert=False,
                multi=True,
            )
        )

    def backwards(self, mongodb_collection, mongodb_database):
        print(
            mongodb_collection.update(
                {},
                {
                    "$rename": {"schedule_desk": "template_desk", "schedule_stage": "template_stage"},
                },
                upsert=False,
                multi=True,
            )
        )



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
from content_api.items.service import ItemsService
from content_api.packages.resource import PackagesResource


logger = logging.getLogger(__name__)


class PackagesService(ItemsService):
    """
    A service that knows how to perform CRUD operations on the `package`
    content types.

    Serves mainly as a proxy to the data layer.
    """

    default_sort = PackagesResource.datasource.get("default_sort", [("versioncreated", -1)])

    def on_fetched_item(self, document):
        """Event handler when a single package is retrieved from database.

        It sets the `uri` field for all associated (referenced) objects.

        :param dict document: fetched MongoDB document representing the package
        """
        self._process_referenced_objects(document)
        super().on_fetched_item(document)

    def on_fetched(self, result):
        """Event handler when a collection of packages is retrieved from
        database.

        For each package in the fetched collection it sets the `uri` field for
        all objects associated with (referenced by) the package.

        :param dict result: dictionary contaning the list of MongoDB documents
            (the fetched packages) and some metadata, e.g. pagination info
        """
        for document in result["_items"]:
            self._process_referenced_objects(document)
        super().on_fetched(result)

    def _process_referenced_objects(self, document):
        """Do some processing on the objects referenced by `document`.

        For all referenced objects their `uri` field is generated and their
        `_id` field removed.

        :param dict document: MongoDB document representing a package object
        """
        for item in document.get("associations", {}).values():
            if not item:
                continue
            if item.get("_id") or item.get("guid"):
                item["uri"] = self._get_uri(item)
                item.pop("_id", None)



from flask import Blueprint, request, Response
import superdesk
from superdesk import get_resource_service
from superdesk.publish.formatters import get_formatter
from apps.content_types import apply_schema


bp = Blueprint("format_document", __name__)


def get_mime_type(formatter_qcode):
    if formatter_qcode == "newsmlg2":
        return "text/xml"
    else:
        return "application/json"


@bp.route("/format-document-for-preview/", methods=["GET", "OPTIONS"])
def format_document():

    document_id = request.args.get("document_id")
    subscriber_id = request.args.get("subscriber_id")
    formatter_qcode = request.args.get("formatter")

    subscriber = get_resource_service("subscribers").find_one(req=None, _id=subscriber_id)
    doc = get_resource_service("archive").find_one(req=None, _id=document_id)

    formatter = get_formatter(formatter_qcode, doc)
    formatted_docs = formatter.format(article=apply_schema(doc), subscriber=subscriber, codes=None)

    headers = {
        "Access-Control-Allow-Origin": "*",
        "Access-Control-Allow-Methods": "GET",
        "Cache-Control": "no-cache, no-store, must-revalidate",
    }

    return Response(formatted_docs[0][1], headers=headers, mimetype=get_mime_type(formatter_qcode))


def init_app(app):

    superdesk.blueprint(bp, app)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json

from unittest.mock import MagicMock
from datetime import timedelta

from eve.versioning import resolve_document_version
from eve.utils import ParsedRequest

from apps.archive.common import insert_into_versions, ARCHIVE
from superdesk import get_resource_service
from superdesk.tests import TestCase
from superdesk.utc import utcnow
from apps.legal_archive.commands import LegalArchiveImport


class LegalArchiveTestCase(TestCase):

    desks = [{"_id": "123", "name": "Sports"}]
    users = [{"_id": "123", "username": "test1", "first_name": "test", "last_name": "user", "email": "a@a.com"}]
    stages = [{"_id": "123", "name": "working stage", "desk": "123"}]
    archive = [
        {"task": {"desk": "123", "stage": "123"}},
        {"task": {"desk": "1234", "stage": None, "user": "123"}},
        {"task": {"desk": "1234", "stage": "dddd", "user": "test"}},
    ]

    def setUp(self):
        self.app.data.insert("desks", self.desks)
        self.app.data.insert("users", self.users)
        self.app.data.insert("stages", self.stages)

    def test_denormalize_desk_user(self):
        LegalArchiveImport()._denormalize_user_desk(self.archive[0], "")
        task = self.archive[0]["task"]
        self.assertEqual(task.get("desk"), "Sports")
        self.assertEqual(task.get("stage"), "working stage")
        self.assertEqual(task.get("user"), "")

    def test_denormalize_not_configured_desk(self):
        LegalArchiveImport()._denormalize_user_desk(self.archive[1], "")
        task = self.archive[1]["task"]
        self.assertEqual(task.get("desk"), "1234")
        self.assertEqual(task.get("stage"), None)
        self.assertEqual(task.get("user"), "test user")

    def test_denormalize_not_configured_desk_stage_user(self):
        LegalArchiveImport()._denormalize_user_desk(self.archive[2], "")
        task = self.archive[2]["task"]
        self.assertEqual(task.get("desk"), "1234")
        self.assertEqual(task.get("stage"), "dddd")
        self.assertEqual(task.get("user"), "")


class ImportLegalArchiveCommandTestCase(TestCase):
    desks = [{"name": "Sports"}]
    users = [{"username": "test1", "first_name": "test", "last_name": "user", "email": "a@a.com"}]

    def setUp(self):
        try:
            from apps.legal_archive.commands import ImportLegalArchiveCommand
        except ImportError:
            self.fail("Could not import class under test (ImportLegalArchiveCommand).")
        else:
            self.class_under_test = ImportLegalArchiveCommand
            self.app.data.insert("desks", self.desks)
            self.app.data.insert("users", self.users)
            self.validators = [
                {"schema": {}, "type": "text", "act": "publish", "_id": "publish_text"},
                {"schema": {}, "type": "text", "act": "correct", "_id": "correct_text"},
                {"schema": {}, "type": "text", "act": "kill", "_id": "kill_text"},
            ]

            self.products = [
                {"_id": "1", "name": "prod1"},
                {"_id": "2", "name": "prod2", "codes": "abc,def"},
                {"_id": "3", "name": "prod3", "codes": "xyz"},
            ]

            self.subscribers = [
                {
                    "name": "Test",
                    "is_active": True,
                    "subscriber_type": "wire",
                    "email": "test@test.com",
                    "sequence_num_settings": {"max": 9999, "min": 1},
                    "products": ["1"],
                    "destinations": [
                        {
                            "name": "test",
                            "delivery_type": "email",
                            "format": "nitf",
                            "config": {"recipients": "test@test.com"},
                        }
                    ],
                }
            ]
            self.app.data.insert("validators", self.validators)
            self.app.data.insert("products", self.products)
            self.app.data.insert("subscribers", self.subscribers)
            self.class_under_test = ImportLegalArchiveCommand
            self.archive_items = [
                {
                    "task": {"desk": self.desks[0]["_id"], "stage": self.desks[0]["incoming_stage"], "user": "123"},
                    "_id": "item1",
                    "state": "in_progress",
                    "headline": "item 1",
                    "type": "text",
                    "slugline": "item 1 slugline",
                    "_current_version": 1,
                    "_created": utcnow() - timedelta(minutes=3),
                    "expired": utcnow() - timedelta(minutes=30),
                },
                {
                    "task": {"desk": self.desks[0]["_id"], "stage": self.desks[0]["incoming_stage"], "user": "123"},
                    "_id": "item2",
                    "state": "in_progress",
                    "headline": "item 2",
                    "type": "text",
                    "slugline": "item 2 slugline",
                    "_current_version": 1,
                    "_created": utcnow() - timedelta(minutes=2),
                    "expired": utcnow() - timedelta(minutes=30),
                },
                {
                    "task": {"desk": self.desks[0]["_id"], "stage": self.desks[0]["incoming_stage"], "user": "123"},
                    "_id": "item3",
                    "state": "in_progress",
                    "headline": "item 2",
                    "type": "text",
                    "slugline": "item 2 slugline",
                    "_current_version": 1,
                    "_created": utcnow() - timedelta(minutes=1),
                    "expired": utcnow() - timedelta(minutes=30),
                },
            ]

            get_resource_service(ARCHIVE).post(self.archive_items)
            for item in self.archive_items:
                resolve_document_version(item, ARCHIVE, "POST")
                insert_into_versions(id_=item["_id"])

    def test_import_into_legal_archive(self):
        archive_publish = get_resource_service("archive_publish")
        archive_correct = get_resource_service("archive_correct")
        legal_archive = get_resource_service("legal_archive")
        archive = get_resource_service("archive_publish")
        published = get_resource_service("published")
        publish_queue = get_resource_service("publish_queue")

        self.original_method = LegalArchiveImport.upsert_into_legal_archive
        LegalArchiveImport.upsert_into_legal_archive = MagicMock()

        for item in self.archive_items:
            archive_publish.patch(item["_id"], {"headline": "publishing", "abstract": "publishing"})

        for item in self.archive_items:
            legal_item = legal_archive.find_one(req=None, _id=item["_id"])
            self.assertIsNone(legal_item, "Item: {} is not none.".format(item["_id"]))

        archive_correct.patch(self.archive_items[1]["_id"], {"headline": "correcting", "abstract": "correcting"})

        LegalArchiveImport.upsert_into_legal_archive = self.original_method
        self.class_under_test().run(1)

        # items are not expired
        for item in self.archive_items:
            legal_item = legal_archive.find_one(req=None, _id=item["_id"])
            self.assertIsNone(legal_item, "Item: {} is not none.".format(item["_id"]))

        # expire the items
        for item in self.archive_items:
            original = archive.find_one(req=None, _id=item["_id"])
            archive.system_update(item["_id"], {"expiry": utcnow() - timedelta(minutes=30)}, original)
            published.update_published_items(item["_id"], "expiry", utcnow() - timedelta(minutes=30))

        # run the command after expiry
        self.class_under_test().run(1)

        # items are expired
        for item in self.archive_items:
            legal_item = legal_archive.find_one(req=None, _id=item["_id"])
            self.assertEqual(legal_item["_id"], item["_id"], "item {} not imported to legal".format(item["_id"]))

        # items are moved to legal
        for item in self.archive_items:
            published_items = list(published.get_other_published_items(item["_id"]))
            for published_item in published_items:
                self.assertEqual(published_item["moved_to_legal"], True)

        # items are moved to legal publish queue
        for item in self.archive_items:
            req = ParsedRequest()
            req.where = json.dumps({"item_id": item["_id"]})
            queue_items = list(publish_queue.get(req=req, lookup=None))
            self.assertGreaterEqual(len(queue_items), 1)
            for queue_item in queue_items:
                self.assertEqual(queue_item["moved_to_legal"], True)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import os
from flask import json
from copy import deepcopy

from apps.prepopulate.app_populate import AppPopulateCommand
from apps.prepopulate.app_initialize import AppInitializeWithDataCommand
from superdesk import tests
from superdesk.factory.app import get_app
from superdesk.tests import setup_auth_user
from superdesk.tests.mocks import TestSearchProvider
from superdesk.tests.steps import get_macro_path
from superdesk.tests.setup_teardown import setup_providers, teardown_providers


readonly_fields = ["display_name", "password", "phone", "first_name", "last_name"]
LDAP_SERVER = os.environ.get("LDAP_SERVER")


def setup_before_all(context, config, app_factory):
    """
    Keep it for backwards compatibility.

    TODO: it needs to be cleaned on each superdesk repo.
    """
    setup_before_all.config = config
    setup_before_all.app_factory = app_factory


def setup_before_scenario(context, scenario, config, app_factory):
    if scenario.status != "skipped" and "notesting" in scenario.tags:
        config["SUPERDESK_TESTING"] = False

    tests.setup(context, config, app_factory, bool(config))

    context.headers = [("Content-Type", "application/json"), ("Origin", "localhost")]

    if "dbauth" in scenario.tags and LDAP_SERVER:
        scenario.mark_skipped()

    if "ldapauth" in scenario.tags and not LDAP_SERVER:
        scenario.mark_skipped()

    if "alchemy" in scenario.tags and not context.app.config.get("KEYWORDS_KEY_API"):
        scenario.mark_skipped()

    if "clean_snapshots" in scenario.tags:
        tests.use_snapshot.cache.clear()

    setup_search_provider(context.app)

    if scenario.status != "skipped" and "auth" in scenario.tags:
        setup_auth_user(context)

    if scenario.status != "skipped" and "provider" in scenario.tags:
        setup_providers(context)

    if scenario.status != "skipped" and "vocabulary" in scenario.tags:
        with context.app.app_context():
            cmd = AppPopulateCommand()
            filename = os.path.join(os.path.abspath(os.path.dirname("features/steps/fixtures/")), "vocabularies.json")
            cmd.run(filename)

    if scenario.status != "skipped" and "content_type" in scenario.tags:
        with context.app.app_context():
            cmd = AppPopulateCommand()
            filename = os.path.join(os.path.abspath(os.path.dirname("features/steps/fixtures/")), "content_types.json")
            cmd.run(filename)

    if scenario.status != "skipped" and "notification" in scenario.tags:
        tests.setup_notification(context)

    if scenario.status != "skipped" and "app_init" in scenario.tags:
        with context.app.app_context():
            command = AppInitializeWithDataCommand()
            command.run()


def before_all(context):
    # https://pythonhosted.org/behave/api.html#logging-setup
    context.config.setup_logging()


def before_feature(context, feature):
    config = getattr(setup_before_all, "config", None)
    if config is not None:
        app_factory = setup_before_all.app_factory
    else:
        # superdesk-aap don't use "setup_before_all" already
        config = getattr(setup_before_scenario, "config", None)
        app_factory = getattr(setup_before_scenario, "app_factory", None)
    config = deepcopy(config or {})
    app_factory = app_factory or get_app

    # set the MAX_TRANSMIT_RETRY_ATTEMPT to zero so that transmit does not retry
    config["MAX_TRANSMIT_RETRY_ATTEMPT"] = 0
    os.environ["BEHAVE_TESTING"] = "1"
    tests.setup(context, config, app_factory=app_factory)

    if "tobefixed" in feature.tags:
        feature.mark_skipped()

    if "dbauth" in feature.tags and LDAP_SERVER:
        feature.mark_skipped()

    if "ldapauth" in feature.tags and not LDAP_SERVER:
        feature.mark_skipped()


def before_scenario(context, scenario):
    config = {}
    setup_before_scenario(context, scenario, config, app_factory=get_app)


def after_scenario(context, scenario):
    if "provider" in scenario.tags:
        teardown_providers(context)

    if "notification" in scenario.tags:
        tests.teardown_notification(context)

    if "clean" in scenario.tags:
        try:
            os.remove(get_macro_path("behave_macro.py"))
            os.remove(get_macro_path("validate_headline_macro.py"))
        except Exception:
            pass


def before_step(context, step):
    if LDAP_SERVER and step.text:
        try:
            step_text_json = json.loads(step.text)
            step_text_json = (
                {k: step_text_json[k] for k in step_text_json.keys() if k not in readonly_fields}
                if isinstance(step_text_json, dict)
                else [
                    {k: json_obj[k] for k in json_obj.keys() if k not in readonly_fields} for json_obj in step_text_json
                ]
            )

            step.text = json.dumps(step_text_json)
        except Exception:
            pass


def setup_search_provider(app):
    from apps.search_providers import register_search_provider, allowed_search_providers

    if "testsearch" not in allowed_search_providers:
        register_search_provider("testsearch", provider_class=TestSearchProvider)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import arrow
import magic
import base64
import hashlib
import logging
import requests
import requests.exceptions
import os

from urllib.parse import urljoin
from bson import ObjectId
from io import BytesIO
from PIL import Image, ImageEnhance
from flask import json, url_for
from .image import get_meta
from .video import get_meta as video_meta
from superdesk.errors import SuperdeskApiError
from flask import current_app as app
from mimetypes import guess_extension

logger = logging.getLogger(__name__)


def hash_file(afile, hasher, blocksize=65536):
    buf = afile.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = afile.read(blocksize)
    return hasher.hexdigest()


def get_file_name(file):
    return hash_file(file, hashlib.sha256())


def fix_content_type(content_type, content):
    """Get known content_type

    It will try to get it from content if it starts with application or is unknown.
    """
    if not content_type or "application/" in content_type:
        content_type = magic.from_buffer(content.getvalue(), mime=True)
        content.seek(0)
    return str(content_type)


def download_file_from_url(url, request_kwargs=None):
    """Download file from given url.

    In case url is relative it will prefix it with current host.

    :param url: file url
    """

    if not request_kwargs:
        request_kwargs = {}

    try:
        rv = requests.get(url, headers={"User-Agent": "Superdesk-1.0"}, timeout=(5, 25), **request_kwargs)
    except requests.exceptions.MissingSchema:  # any route will do here, we only need host
        rv = requests.get(urljoin(url_for("static", filename="x", _external=True), url), timeout=15, **request_kwargs)
    if rv.status_code not in (200, 201):
        raise SuperdeskApiError.internalError("Failed to retrieve file from URL: %s" % url)
    content = BytesIO(rv.content)
    content_type = rv.headers.get("content-type", "image/jpeg").split(";")[0]
    content_type = fix_content_type(content_type, content)
    ext = str(content_type).split("/")[1]
    name = str(ObjectId()) + ext
    return content, name, content_type


def download_file_from_encoded_str(encoded_str):
    content = encoded_str.split(";base64,")
    mime = content[0].split(":")[1]
    ext = content[0].split("/")[1]
    name = str(ObjectId()) + ext
    content = base64.b64decode(content[1])
    return BytesIO(content), name, mime


def process_file_from_stream(content, content_type=None):
    content_type = content_type or content.content_type
    content = BytesIO(content.read())
    content_type = fix_content_type(content_type, content)
    file_type, ext = content_type.split("/")
    try:
        metadata = process_file(content, file_type)
    except OSError:  # error from PIL when image is supposed to be an image but is not.
        raise SuperdeskApiError.internalError("Failed to process file")
    file_name = get_file_name(content)
    content.seek(0)
    metadata = encode_metadata(metadata)
    metadata.update({"length": json.dumps(len(content.getvalue()))})
    return file_name, content_type, metadata


def encode_metadata(metadata):
    return dict((k.lower(), json.dumps(v)) for k, v in metadata.items())


def decode_metadata(metadata):
    return dict((k.lower(), decode_val(v)) for k, v in metadata.items())


def decode_val(string_val):
    """Format dates that elastic will try to convert automatically."""
    val = json.loads(string_val)
    try:
        arrow.get(val, "YYYY-MM-DD")  # test if it will get matched by elastic
        return str(arrow.get(val))
    except (Exception):
        return val


def process_file(content, type):
    """Retrieves the media file metadata

    :param BytesIO content: content stream
    :param str type: type of media file
    :return: dict metadata related to media file.
    """
    if type == "image":
        return process_image(content)
    if type in ("audio", "video"):
        return process_video(content)
    return {}


def process_video(content):
    """Retrieves the video/audio metadata

    :param BytesIO content: content stream
    :return: dict video/audio metadata
    """
    content.seek(0)
    meta = video_meta(content)
    content.seek(0)
    return meta


def process_image(content):
    """Retrieves the image metadata

    :param BytesIO content: content stream
    :return: dict image metadata
    """
    content.seek(0)
    meta = get_meta(content)
    content.seek(0)
    return meta


def _get_cropping_data(doc):
    """Get PIL Image crop data from doc with superdesk crops specs.

    :param doc: crop dict
    """
    if all(
        [
            doc.get("CropTop", None) is not None,
            doc.get("CropLeft", None) is not None,
            doc.get("CropRight", None) is not None,
            doc.get("CropBottom", None) is not None,
        ]
    ):
        return (int(doc["CropLeft"]), int(doc["CropTop"]), int(doc["CropRight"]), int(doc["CropBottom"]))


def crop_image(content, file_name, cropping_data, exact_size=None, image_format=None):
    """Crop image stream to given crop.

    :param content: image file stream
    :param file_name
    :param cropping_data: superdesk crop dict ({'CropLeft': 0, 'CropTop': 0, ...})
    :param exact_size: dict with `width` and `height` values
    """
    if not isinstance(cropping_data, tuple):
        cropping_data = _get_cropping_data(cropping_data)
    if cropping_data:
        logger.debug("Opened image {} from stream, going to crop it".format(file_name))
        content.seek(0)
        img = Image.open(content)
        cropped = img.crop(cropping_data)
        if exact_size and "width" in exact_size and "height" in exact_size:
            cropped = cropped.resize((int(exact_size["width"]), int(exact_size["height"])), Image.ANTIALIAS)
        logger.debug("Cropped image {} from stream, going to save it".format(file_name))
        try:
            out = BytesIO()
            cropped.save(out, image_format or img.format)
            out.seek(0)
            setattr(out, "width", cropped.size[0])
            setattr(out, "height", cropped.size[1])
            return True, out
        except Exception as io:
            logger.exception("Failed to generate crop for filename: {}. Crop: {}".format(file_name, cropping_data))
            return False, io
    return False, content


def get_watermark(image):
    """Apply the specified watermark to the image if available

    :param image:
    :return: watermarked image
    """
    image = image.copy()
    if not app.config.get("WATERMARK_IMAGE"):
        return image
    if image.mode != "RGBA":
        image = image.convert("RGBA")
    path = os.path.join(app.config["ABS_PATH"], app.config["WATERMARK_IMAGE"])
    if not os.path.isfile(path):
        logger.warning("No water mark file found at : {}".format(path))
        return image
    with open(path, mode="rb") as watermark_binary:
        watermark_image = Image.open(watermark_binary)
        set_opacity(watermark_image, 0.3)
        watermark_layer = Image.new("RGBA", image.size)
        watermark_layer.paste(
            watermark_image,
            (
                image.size[0] - watermark_image.size[0],
                int((image.size[1] - watermark_image.size[1]) * 0.66),
            ),
        )

    watermark = Image.alpha_composite(image, watermark_layer)
    return watermark.convert("RGB")


def set_opacity(image, opacity=1):
    alpha = image.split()[3]
    alpha = ImageEnhance.Brightness(alpha).enhance(opacity)
    image.putalpha(alpha)


def guess_media_extension(content_type):
    ext = str(guess_extension(content_type))
    if ext in [".jpe", ".jpeg"]:
        return ".jpg"
    if "mp3" in content_type or "audio/mpeg" in content_type:
        return ".mp3"
    if "flac" in content_type:
        return ".flac"
    return ext if ext != "None" else ""



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license
import logging
import superdesk
from superdesk.celery_app import celery
from superdesk.lock import lock, unlock
from .saved_searches import (
    SavedSearchesService,
    SavedSearchesResource,
    AllSavedSearchesResource,
    SavedSearchItemsResource,
    SavedSearchItemsService,
    AllSavedSearchesService,
)
from superdesk import get_resource_service
from superdesk import es_utils
import pytz
from croniter import croniter
from datetime import datetime
from flask import render_template
from superdesk import emails
import json

logger = logging.getLogger(__name__)
REPORT_SOFT_LIMIT = 60 * 5
LOCK_NAME = "saved_searches_report"


def init_app(app):
    endpoint_name = "saved_searches"
    service = SavedSearchesService(endpoint_name, backend=superdesk.get_backend())
    SavedSearchesResource(endpoint_name, app=app, service=service)

    endpoint_name = "all_saved_searches"
    service = AllSavedSearchesService(endpoint_name, backend=superdesk.get_backend())
    AllSavedSearchesResource(endpoint_name, app=app, service=service)

    endpoint_name = "saved_search_items"
    service = SavedSearchItemsService(endpoint_name, backend=superdesk.get_backend())
    SavedSearchItemsResource(endpoint_name, app=app, service=service)

    superdesk.privilege(
        name="use_global_saved_searches", label="Global searches", description="Use global saved searches."
    )
    superdesk.privilege(
        name="global_saved_searches",
        label="Manage Global Saved Searches",
        description="User can manage other users' global saved searches",
    )
    superdesk.privilege(
        name="saved_searches", label="Manage Saved Searches", description="User can manage saved searches"
    )
    superdesk.privilege(
        name="saved_searches_subscriptions",
        label="Manage Saved Searches Subscriptions",
        description="User can (un)subscribe to saved searches",
    )
    superdesk.privilege(
        name="saved_searches_subscriptions_admin",
        label="Manage Saved Searches Subscriptions For Other Users",
        description="User manage other users saved searches subscriptions ",
    )


def get_next_date(scheduling, base=None):
    """Get next schedule date

    :param str scheduling: task schedule, using cron syntax
    :return datetime: date of next schedule
    """
    if base is None:
        tz = pytz.timezone(superdesk.app.config["DEFAULT_TIMEZONE"])
        base = datetime.now(tz=tz)
    cron_iter = croniter(scheduling, base)
    return cron_iter.get_next(datetime)


def send_report_email(user_id, search, docs):
    """Send saved search report by email.

    :param dict search: saved search data
    :param list found_items: items matching the search request
    """
    users_service = get_resource_service("users")
    user_data = next(users_service.find({"_id": user_id}))
    recipients = [user_data["email"]]
    app = superdesk.app
    admins = app.config["ADMINS"]
    subject = "Saved searches report"
    context = {
        "app_name": app.config["APPLICATION_NAME"],
        "search": search,
        "docs": docs,
        "client_url": app.config["CLIENT_URL"].rstrip("/"),
    }
    text_body = render_template("saved_searches_report.txt", **context)
    html_body = render_template("saved_searches_report.html", **context)
    emails.send_email.delay(
        subject=subject, sender=admins[0], recipients=recipients, text_body=text_body, html_body=html_body
    )


def publish_report(user_id, search_data):
    """Create report for a search and send it by email"""
    search_filter = json.loads(search_data["filter"])
    query = es_utils.filter2query(search_filter, user_id=user_id)
    repos = es_utils.filter2repos(search_filter) or es_utils.REPOS.copy()
    docs = list(superdesk.app.data.elastic.search(query, repos))
    send_report_email(user_id, search_data, docs)


def process_subscribers(subscribers, search, now, isDesk=False):
    do_update = False

    for subscriber_data in subscribers:
        scheduling = subscriber_data["scheduling"]
        next_report = subscriber_data.get("next_report")
        if next_report is None:
            subscriber_data["next_report"] = get_next_date(scheduling)
            do_update = True
        elif next_report <= now:
            if isDesk:
                desk = get_resource_service("desks").find_one(req=None, _id=subscriber_data["desk"])
                for member in (desk or {}).get("members", []):
                    publish_report(member.get("user"), search)
            else:
                publish_report(subscriber_data["user"], search)
            subscriber_data["last_report"] = now
            subscriber_data["next_report"] = get_next_date(scheduling)
            do_update = True

    return do_update


@celery.task(soft_time_limit=REPORT_SOFT_LIMIT)
def report():
    """Check all saved_searches with subscribers, and publish reports"""
    if not lock(LOCK_NAME, expire=REPORT_SOFT_LIMIT + 10):
        return
    try:
        saved_searches = get_resource_service("saved_searches")
        subscribed_searches = saved_searches.find({"subscribers": {"$exists": 1}})
        tz = pytz.timezone(superdesk.app.config["DEFAULT_TIMEZONE"])
        now = datetime.now(tz=tz)
        for search in subscribed_searches:
            do_update = False

            subscribed_users = search["subscribers"].get("user_subscriptions", [])
            try:
                do_update = process_subscribers(subscribed_users, search, now, False) or do_update
            except Exception as e:
                logger.error(
                    "Can't do saved search report for users:\nexception: {e}\ndata: {search}".format(e=e, search=search)
                )

            subscribed_desks = search["subscribers"].get("desk_subscriptions", [])
            try:
                do_update = process_subscribers(subscribed_desks, search, now, True) or do_update
            except Exception as e:
                logger.error(
                    "Can't do saved search report for desks:\nexception: {e}\ndata: {search}".format(e=e, search=search)
                )

            if do_update:
                updates = {"subscribers": search["subscribers"]}
                saved_searches.system_update(search["_id"], updates, search)
    except Exception as e:
        logger.error("Can't report saved searches: {reason}".format(reason=e))
    finally:
        unlock(LOCK_NAME)



import superdesk

from superdesk import get_backend
from apps.highlights.resource import HighlightsResource, MarkedForHighlightsResource
from apps.highlights.service import HighlightsService, MarkedForHighlightsService
from .generate import GenerateHighlightsService, GenerateHighlightsResource


def init_app(app):
    endpoint_name = "highlights"
    service = HighlightsService(endpoint_name, backend=get_backend())
    HighlightsResource(endpoint_name, app=app, service=service)

    endpoint_name = "marked_for_highlights"
    service = MarkedForHighlightsService(endpoint_name, backend=get_backend())
    MarkedForHighlightsResource(endpoint_name, app=app, service=service)

    endpoint_name = "generate_highlights"
    service = GenerateHighlightsService(endpoint_name, backend=get_backend())
    GenerateHighlightsResource(endpoint_name, app=app, service=service)

    superdesk.privilege(name="highlights_read", label="Highlights - read", description="Can see a list of highlights.")
    superdesk.privilege(
        name="highlights",
        label="Highlights/Summary List Management",
        description="User can manage highlights/summary lists.",
    )
    superdesk.privilege(
        name="mark_for_highlights",
        label="Mark items for Highlights/Summary Lists",
        description="User can mark items for Highlights/Summary Lists.",
    )



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
import copy
import logging
from flask_babel import _
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.utils import ListCursor
from superdesk import get_resource_service, config, app
from superdesk.io.subjectcodes import get_subjectcodeitems
from eve.utils import ParsedRequest


logger = logging.getLogger(__name__)


class FilterConditionParametersResource(Resource):
    url = "filter_conditions/parameters"
    resource_methods = ["GET"]
    item_methods = []
    schema = {
        "field": {"type": "string"},
        "label": {"type": "string"},
        "operators": {"type": "list"},
        "values": {"type": "list"},
        "value_field": {"type": "string"},
    }


class FilterConditionParametersService(BaseService):
    def get(self, req, lookup):
        values = self._get_field_values()
        fields = [
            {
                "field": "anpa_category",
                "label": _("ANPA Category"),
                "operators": ["in", "nin"],
                "values": values.get("anpa_category", []),
                "value_field": "qcode",
            },
            {
                "field": "urgency",
                "label": _("Urgency"),
                "operators": ["in", "nin", "eq", "ne", "lt", "lte", "gt", "gte"],
                "values": values.get("urgency", []),
                "value_field": "qcode",
            },
            {
                "field": "genre",
                "label": _("Genre"),
                "operators": ["in", "nin"],
                "values": values.get("genre", []),
                "value_field": "qcode",
            },
            {
                "field": "subject",
                "label": _("Subject"),
                "operators": ["in", "nin"],
                "values": values.get("subject", []),
                "value_field": "qcode",
            },
            {
                "field": "priority",
                "label": _("Priority"),
                "operators": ["in", "nin", "eq", "ne", "lt", "lte", "gt", "gte"],
                "values": values.get("priority", []),
                "value_field": "qcode",
            },
            {"field": "keywords", "label": _("Keywords"), "operators": ["in", "nin"]},
            {
                "field": "slugline",
                "label": _("Slugline"),
                "operators": ["in", "nin", "eq", "ne", "like", "notlike", "startswith", "endswith"],
            },
            {
                "field": "type",
                "label": _("Type"),
                "operators": ["in", "nin", "eq", "ne"],
                "values": values.get("type", []),
                "value_field": "qcode",
            },
            {
                "field": "source",
                "label": _("Source"),
                "operators": ["in", "nin", "eq", "ne", "like", "notlike", "startswith", "endswith"],
            },
            {
                "field": "headline",
                "label": _("Headline"),
                "operators": ["in", "nin", "eq", "ne", "like", "notlike", "startswith", "endswith"],
            },
            {
                "field": "ednote",
                "label": _("Ednote"),
                "operators": ["in", "nin", "eq", "ne", "like", "notlike", "startswith", "endswith"],
            },
            {
                "field": "body_html",
                "label": _("Body HTML"),
                "operators": ["in", "nin", "like", "notlike", "startswith", "endswith"],
            },
            {
                "field": "desk",
                "label": _("Desk"),
                "operators": ["in", "nin", "eq", "ne"],
                "values": values["desk"],
                "value_field": "_id",
            },
            {
                "field": "stage",
                "label": _("Stage"),
                "operators": ["in", "nin", "eq", "ne"],
                "values": values["stage"],
                "value_field": "_id",
            },
            {
                "field": "sms",
                "label": _("SMS"),
                "operators": ["in", "nin", "eq", "ne"],
                "values": values["sms"],
                "value_field": "name",
            },
            {
                "field": "place",
                "label": _("Place"),
                "operators": ["match", "in", "nin"],
                "values": values["place"],
                "value_field": "qcode",
            },
            {
                "field": "ingest_provider",
                "label": _("Ingest provider"),
                "operators": ["eq", "ne"],
                "values": values["ingest_provider"],
                "value_field": "_id",
            },
            {
                "field": "embargo",
                "label": _("Embargo"),
                "operators": ["eq", "ne"],
                "values": values["embargo"],
                "value_field": "name",
            },
            {
                "field": "featuremedia",
                "label": _("Feature Media"),
                "operators": ["exists"],
                "values": values["featuremedia"],
                "value_field": "name",
            },
            {
                "field": "anpa_take_key",
                "operators": ["in", "nin", "eq", "ne", "like", "notlike", "startswith", "endswith"],
            },
        ]

        if "planning" in app.config.get("INSTALLED_APPS", []):
            fields.append(
                {
                    "field": "agendas",
                    "label": _("Agendas"),
                    "operators": ["in", "nin"],
                    "values": list(get_resource_service("agenda").find({"is_enabled": True})),
                    "value_field": "_id",
                }
            )

        fields.extend(self._get_vocabulary_fields(values))
        return ListCursor(fields)

    def _get_vocabulary_fields(self, values):
        excluded_vocabularies = copy.copy(app.config.get("EXCLUDED_VOCABULARY_FIELDS", []))
        excluded_vocabularies.extend(values)
        lookup = {"_id": {"$nin": excluded_vocabularies}, "type": "manageable"}
        for vocabulary in get_resource_service("vocabularies").get(req=None, lookup=lookup):
            field = {"field": vocabulary[config.ID_FIELD], "label": vocabulary["display_name"]}

            if vocabulary.get("field_type") and vocabulary.get("field_type", "") != "text":
                continue

            if vocabulary.get("field_type", "") == "text":
                field["operators"] = ["in", "nin", "eq", "ne", "like", "notlike", "startswith", "endswith"]
            else:
                field["values"] = vocabulary["items"]
                field["operators"] = ["in", "nin"]
                field["value_field"] = "qcode"
            yield field

    def _get_field_values(self):
        values = {}
        vocabularies_resource = get_resource_service("vocabularies")
        categories_cv = vocabularies_resource.find_one(req=None, _id="categories")
        values["anpa_category"] = categories_cv.get("items") if categories_cv else []
        req = ParsedRequest()
        req.where = json.dumps({"$or": [{"schema_field": "genre"}, {"_id": "genre"}]})
        genre = vocabularies_resource.get(req=req, lookup=None)
        if genre.count():
            values["genre"] = genre[0]["items"]
        for voc_id in ("urgency", "priority", "type"):
            try:
                values[voc_id] = vocabularies_resource.find_one(req=None, _id=voc_id)["items"]
            except TypeError:
                values[voc_id] = []
        subject = vocabularies_resource.find_one(req=None, schema_field="subject")
        if subject:
            values["subject"] = subject["items"]
        else:
            values["subject"] = get_subjectcodeitems()
        values["desk"] = list(get_resource_service("desks").get(None, {}))
        values["stage"] = self._get_stage_field_values(values["desk"])
        values["sms"] = [{"qcode": 0, "name": "False"}, {"qcode": 1, "name": "True"}]
        values["embargo"] = [{"qcode": 0, "name": "False"}, {"qcode": 1, "name": "True"}]
        req = ParsedRequest()
        req.where = json.dumps({"$or": [{"schema_field": "place"}, {"_id": "place"}, {"_id": "locators"}]})
        place = vocabularies_resource.get(req=req, lookup=None)
        if place.count():
            values["place"] = place[0]["items"]
        else:
            values["place"] = []
        values["ingest_provider"] = list(get_resource_service("ingest_providers").get(None, {}))
        values["featuremedia"] = [{"qcode": 1, "name": "True"}, {"qcode": 0, "name": "False"}]
        return values

    def _get_stage_field_values(self, desks):
        stages = list(get_resource_service("stages").get(None, {}))
        for i, stage in enumerate(stages):
            try:
                desk = next(filter(lambda d: d["_id"] == stage["desk"], desks))
            except (StopIteration, KeyError):
                # if stage has no desk, remove that stage from a list
                logger.warning('Desk not found for stage with id "{}".'.format(stage["_id"]))
                stages[i] = None
                continue
            stages[i]["name"] = "{}: {}".format(desk["name"], stage["name"])
        return tuple(i for i in stages if i)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
from lxml import etree
from superdesk.metadata.item import ITEM_TYPE, CONTENT_TYPE, FORMATS, FORMAT
from superdesk.etree import parse_html
from superdesk.text_utils import get_text
from superdesk.publish import registered_transmitters

formatters = []
logger = logging.getLogger(__name__)


class FormatterRegistry(type):
    """Registry metaclass for formatters."""

    def __init__(cls, name, bases, attrs):
        """Register sub-classes of Formatter class when defined."""
        super(FormatterRegistry, cls).__init__(name, bases, attrs)
        if name != "Formatter":
            formatters.append(cls)


class Formatter(metaclass=FormatterRegistry):
    """Base Formatter class for all types of Formatters like News ML 1.2, News ML G2, NITF, etc."""

    def __init__(self):
        self.can_preview = False
        self.can_export = False
        self.destination = None
        self.subscriber = None

    def format(self, article, subscriber, codes=None):
        """Formats the article and returns the transformed string"""
        raise NotImplementedError()

    def export(self, article, subscriber, codes=None):
        """Formats the article and returns the output string for export"""
        raise NotImplementedError()

    def can_format(self, format_type, article):
        """Test if formatter can format for given article."""
        raise NotImplementedError()

    def append_body_footer(self, article):
        """
        Checks if the article has any Public Service Announcements and if available appends each of them to the body.

        :return: body with public service announcements.
        """
        try:
            article["body_html"] = article["body_html"].replace("<br>", "<br/>")
        except KeyError:
            pass

        body = ""
        if article[ITEM_TYPE] in [CONTENT_TYPE.TEXT, CONTENT_TYPE.PREFORMATTED]:
            body = article.get("body_html", "")
        elif article[ITEM_TYPE] in [CONTENT_TYPE.AUDIO, CONTENT_TYPE.PICTURE, CONTENT_TYPE.VIDEO]:
            body = article.get("description", "")

        if body and article.get(FORMAT, "") == FORMATS.PRESERVED:
            body = body.replace("\n", "\r\n").replace("\r\r", "\r")
            parsed = parse_html(body, content="html")

            for br in parsed.xpath("//br"):
                br.tail = "\r\n" + br.tail if br.tail else "\r\n"

            etree.strip_elements(parsed, "br", with_tail=False)
            body = etree.tostring(parsed, encoding="unicode")

        if body and article.get("body_footer"):
            footer = article.get("body_footer")
            if article.get(FORMAT, "") == FORMATS.PRESERVED:
                body = "{}\r\n{}".format(body, get_text(footer))
            else:
                body = "{}{}".format(body, footer)
        return body

    def append_legal(self, article, truncate=False):
        """
        Checks if the article has the legal flag on and adds 'Legal:' to the slugline

        :param article: article having the slugline
        :param truncate: truncates the slugline to 24 characters
        :return: updated slugline
        """
        slugline = article.get("slugline", "") or ""

        if article.get("flags", {}).get("marked_for_legal", False):
            slugline = "{}: {}".format("Legal", slugline)
            if truncate:
                slugline = slugline[:24]

        return slugline

    def map_html_to_xml(self, element, html):
        """
        Map the html text tags to xml

        :param etree.Element element: The xml element to populate
        :param str html: the html to parse the text from
        :return:
        """
        root = parse_html(html, content="html")
        # if there are no ptags just br
        if not len(root.xpath("//p")) and len(root.xpath("//br")):
            para = etree.SubElement(element, "p")
            for br in root.xpath("//br"):
                etree.SubElement(para, "br").text = br.text

        for p in root.xpath("//p"):
            para = etree.SubElement(element, "p")
            if len(p.xpath(".//br")) > 0:
                for br in p.xpath(".//br"):
                    etree.SubElement(para, "br").text = br.text
            para.text = etree.tostring(p, encoding="unicode", method="text")

        # there neither ptags pr br's
        if len(list(element)) == 0:
            etree.SubElement(element, "p").text = etree.tostring(root, encoding="unicode", method="text")

    def set_destination(self, destination=None, subscriber=None):
        self.destination = destination
        self.subscriber = subscriber

    def _publish_media(self, media):
        if self.destination:
            try:
                transmitter = registered_transmitters[self.destination["delivery_type"]]
            except KeyError:
                logger.warning("Missing transmitter for destination %s", self.destination)
            else:
                return transmitter.transmit_media(media, self.subscriber, self.destination)


def get_formatter(format_type, article):
    for formatter_cls in formatters:
        formatter_instance = formatter_cls()
        if formatter_instance.can_format(format_type, article):
            return formatter_instance


def get_all_formatters():
    """Return all formatters registered."""
    return [formatter_cls() for formatter_cls in formatters]


from .nitf_formatter import NITFFormatter  # NOQA
from .ninjs_formatter import NINJSFormatter, NINJS2Formatter  # NOQA
from .newsml_1_2_formatter import NewsML12Formatter  # NOQA
from .newsml_g2_formatter import NewsMLG2Formatter  # NOQA
from .email_formatter import EmailFormatter  # NOQA
from .ninjs_newsroom_formatter import NewsroomNinjsFormatter  # NOQA
from .idml_formatter import IDMLFormatter  # NOQA
from .ninjs_ftp_formatter import FTPNinjsFormatter  # NOQA
from .imatrics import IMatricsFormatter  # NOQA



import flask
import unittest.mock

from apps.content import push_content_notification


class PushContentNotificationTestCase(unittest.TestCase):
    def setUp(self):
        self.app = flask.Flask(__name__)
        self.ctx = self.app.app_context()
        self.ctx.push()

    def tearDown(self):
        self.ctx.pop()

    @unittest.mock.patch("apps.content.push_notification")
    def test_push_content_notification(self, push_notification):
        foo1 = {"_id": "foo", "task": {"desk": "sports", "stage": "inbox"}}
        foo2 = {"_id": "foo", "task": {"desk": "news", "stage": "todo"}}
        foo3 = {"_id": "foo"}

        push_content_notification([foo1, foo2, foo3])
        push_notification.assert_called_once_with(
            "content:update", user="", items={"foo": 1}, desks={"sports": 1, "news": 1}, stages={"inbox": 1, "todo": 1}
        )



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.resource import Resource
from superdesk.metadata.utils import ProductTypes


class ProductsResource(Resource):
    """Products schema"""

    schema = {
        "name": {"type": "string", "iunique": True, "required": True},
        "description": {"type": "string"},
        "codes": {"type": "string"},
        "content_filter": {
            "type": "dict",
            "schema": {
                "filter_id": Resource.rel("content_filters", nullable=True),
                "filter_type": {"type": "string", "allowed": ["blocking", "permitting"], "default": "blocking"},
            },
            "nullable": True,
        },
        "geo_restrictions": {"type": "string", "nullable": True},
        "product_type": {
            "type": "string",
            "default": ProductTypes.BOTH.value,
            "allowed": ProductTypes.values(),
            "required": True,
        },
        "init_version": {"type": "integer"},
    }

    privileges = {"POST": "products", "PATCH": "products", "DELETE": "products"}

    mongo_indexes = {
        "name_1": ([("name", 1)], {"unique": True}),
    }



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk
from apps.dictionaries.resource import DictionariesResource
from apps.dictionaries.service import DictionaryService


def init_app(app):
    endpoint_name = "dictionaries"
    service = DictionaryService(endpoint_name, backend=superdesk.get_backend())
    DictionariesResource(endpoint_name, app=app, service=service)

    superdesk.privilege(
        name="dictionaries", label="Dictionaries List Management", description="User can manage dictionaries lists."
    )



__all__ = (
    "Schema",
    "NoneField",
    "SchemaField",
    "StringField",
    "IntegerField",
    "ListField",
    "DictField",
    "RelatedContentField",
)


class SchemaType(type):
    """Type for schema.

    It makes Schema class iterable, so you can get the schema dict
    for it using ``dict(Schema)``.
    """

    def __iter__(self):
        for k in dir(self):
            if isinstance(getattr(self, k), SchemaField):
                yield (k, getattr(self, k).schema)


class Schema(metaclass=SchemaType):
    """Base class for schema definition.

    This allows you to define schema as a class with attributes,
    which makes it easier for documentation than normal dictionary.
    """

    pass


class SchemaField:
    """Schema field.

    These are for defining Schema class attributes.
    """

    def __repr__(self):
        return "any"

    def __init__(self):
        self.schema = {}


class NoneField(SchemaField):
    """None field."""

    def __repr__(self):
        return "none"

    def __init__(self):
        self.schema = None


class IntegerField(SchemaField):
    """Integer schema field."""

    def __repr__(self):
        return "integer"

    def __init__(self, required=False):
        super().__init__()
        self.schema["type"] = "integer"
        self.schema["required"] = required


class ListField(SchemaField):
    """List schema field."""

    def __repr__(self):
        return "list"

    def __init__(self, required=False, mandatory_in_list=None, schema=None):
        super().__init__()
        self.schema["type"] = "list"
        self.schema["required"] = required
        self.schema["mandatory_in_list"] = mandatory_in_list
        self.schema["schema"] = schema


class DictField(SchemaField):
    """Dict schema field."""

    def __repr__(self):
        return "dict"

    def __init__(self, required=False, schema=None):
        super().__init__()
        self.schema["type"] = "dict"
        self.schema["required"] = required


class MediaField(SchemaField):
    """Media schema field."""

    def __repr__(self):
        return "media"

    def __init__(self, required=False, schema=None):
        super().__init__()
        self.schema["type"] = "media"
        self.schema["required"] = required


class RelatedContentField(SchemaField):
    """Related Content schema field."""

    def __repr__(self):
        return "related-content"

    def __init__(self, required=False, schema=None):
        super().__init__()
        self.schema["type"] = "related-content"
        self.schema["required"] = required


class StringField(SchemaField):
    """String schema field."""

    def __repr__(self):
        return "string"

    def __init__(self, required=False, maxlength=None, minlength=None):
        super().__init__()
        self.schema["type"] = "string"
        self.schema["required"] = required
        self.schema["minlength"] = minlength
        self.schema["maxlength"] = maxlength



from lxml import etree

from ..base import BasePackageElement


class Story(BasePackageElement):
    """
    Story which represents `p`, `h1`, `h2`, `h3`, `h4`, `h5`, `h6`, `blockquote`, `pre`, `headline`, `byline` html tags.

    As Adobe's idml specification explains:
    ``
    A story, or “text flow” is the basic text container in an InDesign document;
    all text exists inside stories. Stories are associated with at least one text frame or text path,
    and can span any number of linked text frames or text paths in a document.
    ``
    """

    STORY_DEFAULTS = {"AppliedNamedGrid": "n", "AppliedTOCStyle": "n", "TrackChanges": "false", "StoryTitle": "$ID/"}
    STORYPREFERENCE_DEFAULTS = {
        "OpticalMarginAlignment": "false",
        "OpticalMarginSize": "12",
        "FrameType": "TextFrameType",
        "StoryOrientation": "Horizontal",
        "StoryDirection": "LeftToRightDirection",
    }
    HYPERLINKTEXTSOURCE_DEFAULTS = {"Hidden": "false", "AppliedCharacterStyle": "CharacterStyle/$ID/Hyperlink"}
    PARAGRAPHSTYLERANGE_DEFAULTS = {"AppliedParagraphStyle": "ParagraphStyle/$ID/NormalParagraphStyle"}
    CHARACTERSTYLERANGE_DEFAULTS = {
        "AppliedCharacterStyle": "CharacterStyle/$ID/[No character style]",
    }
    BLOCK_TAGS_MAPPING = {
        "p": {
            "markup_tag": "NormalParagraph",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Paragraphs%3aNormalParagraph"},
        },
        "blockquote": {
            "markup_tag": "Blockquote",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Paragraphs%3aBlockquote"},
        },
        "pre": {
            "markup_tag": "Preformatted",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Paragraphs%3aPreformatted"},
        },
        "h1": {
            "markup_tag": "Heading1",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Headings%3aHeading1"},
        },
        "h2": {
            "markup_tag": "Heading2",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Headings%3aHeading2"},
        },
        "h3": {
            "markup_tag": "Heading3",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Headings%3aHeading3"},
        },
        "h4": {
            "markup_tag": "Heading4",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Headings%3aHeading4"},
        },
        "h5": {
            "markup_tag": "Heading5",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Headings%3aHeading5"},
        },
        "h6": {
            "markup_tag": "Heading6",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Headings%3aHeading6"},
        },
        "table": {
            "markup_tag": "Table",
            "Table": {"AppliedTableStyle": "TableStyle/$ID/NormalTable"},
        },
        "ul": {
            "markup_tag": "UnorderedList",
            "ParagraphStyleRange": {
                "AppliedParagraphStyle": "ParagraphStyle/Lists%3aUnorderedList",
                "BulletsAndNumberingListType": "BulletList",
            },
        },
        "ol": {
            "markup_tag": "OrderedList",
            "ParagraphStyleRange": {
                "AppliedParagraphStyle": "ParagraphStyle/Lists%3aOrderedList",
                "BulletsAndNumberingListType": "NumberedList",
            },
        },
        # custom
        "headline": {
            "markup_tag": "Headline",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Custom%3aHeadline"},
        },
        "byline": {
            "markup_tag": "Byline",
            "ParagraphStyleRange": {"AppliedParagraphStyle": "ParagraphStyle/Custom%3aByline"},
        },
    }
    INLINE_TAGS_MAPPING = {
        "b": [{"attribute": "FontStyle", "value": "Bold", "appendable": True}],
        "i": [{"attribute": "FontStyle", "value": "Italic", "appendable": True}],
        "u": [{"attribute": "Underline", "value": "true", "appendable": False}],
        "s": [{"attribute": "StrikeThru", "value": "true", "appendable": False}],
        "sub": [{"attribute": "Position", "value": "Subscript", "appendable": False}],
        "sup": [{"attribute": "Position", "value": "Superscript", "appendable": False}],
        "a": [{"attribute": "AppliedCharacterStyle", "value": "CharacterStyle/$ID/Hyperlink", "appendable": False}],
    }

    def __init__(self, self_id, element, attributes=None, markup_tag=None):
        self.self_id = self_id
        self._element = element
        self._markup_tag = markup_tag
        self.links = []
        super().__init__(attributes)

    @property
    def filename(self):
        """
        Filename inside IDML package.
        Used as a filename for a file inside zip container.
        :return str: filename
        """
        return "Stories/Story_{}.xml".format(self.self_id)

    def _build_etree(self):
        self._etree = etree.Element(etree.QName(self.XMLNS_IDPKG, "Story"), nsmap={"idPkg": self.XMLNS_IDPKG})
        self._etree.set("DOMVersion", self.DOM_VERSION)
        self._add_story()

    def _add_story(self):
        """
        Create <Story..> tag
        :return lxml.etree: Story etree instance
        """
        # merge Story attributes
        story_attributes = self.merge_attributes(self.STORY_DEFAULTS, self._attributes.get("Story", {}))
        story_attributes.update({"Self": self.self_id})
        # Story
        story = etree.SubElement(self._etree, "Story", attrib=story_attributes)
        # StoryPreference
        etree.SubElement(
            story,
            "StoryPreference",
            attrib=self.merge_attributes(self.STORYPREFERENCE_DEFAULTS, self._attributes.get("StoryPreference", {})),
        )

        if self._markup_tag:
            # XMLElement to tag a story
            paragraphstylerange_container = etree.SubElement(
                story,
                "XMLElement",
                attrib={
                    "Self": "{}_{}".format(self.self_id, self._markup_tag.lower()),
                    "XMLContent": self.self_id,
                    "MarkupTag": "XMLTag/{}".format(self._markup_tag),
                },
            )
        else:
            paragraphstylerange_container = story

        # ParagraphStyleRange
        paragraphstylerange = etree.SubElement(
            paragraphstylerange_container,
            "ParagraphStyleRange",
            attrib=self.merge_attributes(
                self.PARAGRAPHSTYLERANGE_DEFAULTS, self._attributes.get("ParagraphStyleRange", {})
            ),
        )

        paragraphstylerange[:] = self._handle_inline_tags(self._element)

        return story

    def _handle_inline_tags(self, element):
        """
        Convert html element's inline style to idml representation.
        :param lxml.etree element: html element
        :return list: list of CharacterStyleRange lxml.etree instances
        """
        element_dict = self._etree_element_to_dict(element)
        return self._create_characterstylerange_recursive(element_dict)

    def _etree_element_to_dict(self, element, parents=None):
        """
        Converts html element to dict.
        Example:
            `<p>normal <b>bold <u>bold-unerline</u> bold</b> normal</p>`
            will be converted to:
            ``
            {
                'attrib': {},
                'childs': (
                    {
                        'attrib': {},
                        'childs': (
                            {
                                'attrib': {},
                                'childs': (),
                                'parents': ('p', 'b'),
                                'tag': 'u',
                                'tail': ' bold',
                                'text': 'bold-unerline',
                            },
                        ),
                        'parents': ('p', ),
                        'tag': 'b',
                        'tail': ' normal',
                        'text': 'bold ',
                    },
                ),
                'parents': (),
                'tag': 'p',
                'tail': None,
                'text': 'normal ',
            }
            ``
        :param lxml.etree element: html element
        :param list parents: list of parent tags
        :return dict: dictionary

        """
        if not parents:
            parents = []

        element_dict = {
            "tag": element.tag,
            "attrib": element.attrib,
            "parents": tuple(parents),
            "text": element.text,
            "tail": element.tail,
        }
        parents.append(element.tag)
        element_dict["childs"] = tuple(
            self._etree_element_to_dict(child, list(parents)) for child in element.iterchildren()
        )

        return element_dict

    def _create_characterstylerange_recursive(self, element_dict, tail=False):
        """
        Create CharacterStyleRange(s) recursively.
        Example:
        ``
        <ParagraphStyleRange ...>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold">
          <Content>Curabitur </Content>
        </CharacterStyleRange>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold Italic">
          <Content>arcu erat</Content>
        </CharacterStyleRange>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold">
          <Content>, accumsan </Content>
        </CharacterStyleRange>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold" Underline="true">
          <Content>id imperdiet et</Content>
        </CharacterStyleRange>
        ...
        </ParagraphStyleRange>
        ``
        :param dict element_dict: dict representation of html element
        :param bool tail: if element has tail
        :return:

        """

        characterstylerange_list = []

        def apply_tag_style(characterstylerange, tag):
            for rule in self.INLINE_TAGS_MAPPING[tag]:
                attribute = rule["attribute"]
                value = rule["value"]
                appendable = rule["appendable"]

                if characterstylerange.get(attribute) and appendable:
                    if attribute == "FontStyle" and value == "Bold":
                        characterstylerange.set(attribute, "{} {}".format(value, characterstylerange.get(attribute)))
                    else:
                        characterstylerange.set(attribute, "{} {}".format(characterstylerange.get(attribute), value))
                else:
                    characterstylerange.set(attribute, value)

        # text
        if element_dict["text"] and len(element_dict["text"]):
            characterstylerange_list.append(
                etree.Element("CharacterStyleRange", attrib=self.CHARACTERSTYLERANGE_DEFAULTS)
            )
            # apply parents styles
            for tag in element_dict["parents"]:
                if self.INLINE_TAGS_MAPPING.get(tag):
                    apply_tag_style(characterstylerange_list[-1], tag)

            # apply own style
            tag = element_dict["tag"]
            if self.INLINE_TAGS_MAPPING.get(tag):
                apply_tag_style(characterstylerange_list[-1], tag)

            content = etree.Element("Content")
            content.text = element_dict["text"]

            if tag == "a":
                self.links.append(
                    {
                        "self_id": "{}_{}".format(self.self_id, len(self.links) + 1),
                        "text": element_dict["text"],
                        "href": element_dict["attrib"].get("href"),
                    }
                )

                hyperlinktextsource = etree.SubElement(
                    characterstylerange_list[-1],
                    "HyperlinkTextSource",
                    attrib=self.merge_attributes(
                        self.HYPERLINKTEXTSOURCE_DEFAULTS,
                        {
                            "Self": self.links[-1]["self_id"],
                            "Name": "Hyperlink {}".format(self.links[-1]["self_id"]),
                        },
                    ),
                )
                hyperlinktextsource.append(content)
            else:
                characterstylerange_list[-1].append(content)

        # childs
        for child in element_dict["childs"]:
            characterstylerange_list += self._create_characterstylerange_recursive(child, tail=True)

        # tail
        if tail and element_dict["tail"] and len(element_dict["tail"]):
            characterstylerange_list.append(
                etree.Element("CharacterStyleRange", attrib=self.CHARACTERSTYLERANGE_DEFAULTS)
            )
            # apply parents styles
            for tag in element_dict["parents"]:
                if self.INLINE_TAGS_MAPPING.get(tag):
                    apply_tag_style(characterstylerange_list[-1], tag)

            content = etree.SubElement(characterstylerange_list[-1], "Content")
            content.text = element_dict["tail"]

        return characterstylerange_list

    @property
    def length(self):
        return len(" ".join(etree.XPath(".//text()")(self._element)))

    @staticmethod
    def guess_height(story, inner_width):
        style = story._etree.xpath(".//ParagraphStyleRange")[-1].get("AppliedParagraphStyle")
        if not style:
            return 100

        style = style.rsplit("%3a", 1)[-1]
        style_map = {
            "Headline": {"PointSize": 48, "Leading": 48, "LeftIndent": 0, "RightIndent": 0},
            "Byline": {"PointSize": 20, "Leading": 20, "LeftIndent": 0, "RightIndent": 0},
            "Heading1": {"PointSize": 40, "Leading": 40, "LeftIndent": 0, "RightIndent": 0},
            "Heading2": {"PointSize": 30, "Leading": 30, "LeftIndent": 0, "RightIndent": 0},
            "Heading3": {"PointSize": 20, "Leading": 20, "LeftIndent": 0, "RightIndent": 0},
            "Heading4": {"PointSize": 14, "Leading": 14, "LeftIndent": 0, "RightIndent": 0},
            "Heading5": {"PointSize": 11, "Leading": 11, "LeftIndent": 0, "RightIndent": 0},
            "Heading6": {"PointSize": 9, "Leading": 9, "LeftIndent": 0, "RightIndent": 0},
            "Blockquote": {"PointSize": 12, "Leading": 12, "LeftIndent": 30, "RightIndent": 30},
            "default": {"PointSize": 12, "Leading": 12, "LeftIndent": 0, "RightIndent": 0},
        }

        point_size = style_map.get(style, style_map["default"])["PointSize"]
        leading = style_map.get(style, style_map["default"])["Leading"]
        left_indent = style_map.get(style, style_map["default"])["LeftIndent"]
        right_indent = style_map.get(style, style_map["default"])["RightIndent"]

        length = story.length
        inner_width -= float(left_indent) + float(right_indent)
        length *= float(point_size) / style_map["default"]["PointSize"]
        length *= float(leading) / style_map["default"]["Leading"]
        height = length / inner_width * 75 + 10

        return height
